from collections import namedtuple
from enum import Enum
from typing import Sequence, Union

from webauthn_rp.errors import UnimplementedError, ValidationError
from webauthn_rp.types import COSEKeyType, CredentialPublicKey


def validate_kty(validator_ktys: Union[str, Sequence[str]],
                 credential_public_key: CredentialPublicKey):
  validator_ktys: Sequence[str] = [validator_ktys] if (  # type: ignore
      type(validator_ktys) is str) else validator_ktys

  for kty in validator_ktys:
    if credential_public_key.kty.name == kty: return

  raise ValidationError(
      'Invalid public key alg type {}, expecting one of {}'.format(
          credential_public_key.kty.name, validator_ktys))


def validate_key_ops(validator_key_ops: Union[str, Sequence[str]],
                     credential_public_key: CredentialPublicKey):
  validator_key_ops: Sequence[str] = [validator_key_ops] if (  # type: ignore
      type(validator_key_ops) is str) else validator_key_ops
  if len(validator_key_ops) == 0: return
  if credential_public_key.key_ops is None:
    raise ValidationError('Found no key ops')

  ops = set(validator_key_ops)
  for kop in credential_public_key.key_ops:
    if kop.name in ops: ops.remove(kop.name)

  if ops:
    raise ValidationError('Missing key ops {}'.format(list(ops)))


def ecdsa_validator(credential_public_key: CredentialPublicKey,
                    sign: bool = False,
                    verify: bool = False):
  validate_kty('EC2', credential_public_key)
  ops = []
  if sign: ops.append('SIGN')
  if verify: ops.append('VERIFY')
  validate_key_ops(ops, credential_public_key)


def eddsa_validator(credential_public_key: CredentialPublicKey,
                    sign: bool = False,
                    verify: bool = False):
  validate_kty('OKP', credential_public_key)
  ops = []
  if sign: ops.append('SIGN')
  if verify: ops.append('VERIFY')
  validate_key_ops(ops, credential_public_key)


class CredentialPublicKeyValidator(Enum):
  ES256 = ecdsa_validator
  ES384 = ecdsa_validator
  ES512 = ecdsa_validator

  EDDSA = eddsa_validator


def validate(credential_public_key: CredentialPublicKey):
  if credential_public_key.alg is None: return
  try:
    validator = getattr(CredentialPublicKeyValidator,
                        credential_public_key.alg.name)
  except AttributeError:
    raise UnimplementedError(
        'Unsupported credential public key alg enum name {}'.format(
            credential_public_key.alg.name))

  validator(credential_public_key)



import base64
import re
from urllib.parse import urlparse

from webauthn_rp.errors import ValidationError


def snake_to_camel_case(s: str) -> str:
  chunks = [x for x in re.split(r'_+', s) if x]
  capped = [x[0].upper() + x[1:] for x in chunks[1:]]
  if chunks:
    return chunks[0] + ''.join(capped)
  return ''


def camel_to_snake_case(s: str) -> str:
  words = []
  s_index = 0
  for i in range(len(s)):
    if s[i].isupper():
      words.append(s[s_index:i].lower())
      s_index = i
  if s_index < len(s): words.append(s[s_index:].lower())
  return '_'.join(words)


def url_base64_encode(b: bytes) -> bytes:
  return base64.b64encode(b, b'-_')


def url_base64_decode(s: str) -> bytes:
  return base64.b64decode(s + '===', b'-_')


def extract_origin(url: str) -> str:
  parsed_url = urlparse(url)
  if parsed_url.netloc is None:
    raise ValidationError('Origin must contain hostname')

  if parsed_url.scheme is None:
    raise ValidationError('Origin must contain scheme')

  return parsed_url.scheme + '://' + parsed_url.netloc



class WebAuthnRPError(Exception):
  pass


class DecodingError(WebAuthnRPError):
  pass


class ParseError(WebAuthnRPError):
  pass


class IntegrityError(WebAuthnRPError):
  pass


class VerificationError(WebAuthnRPError):
  pass


class AuthenticationError(WebAuthnRPError):
  pass


class TokenBindingError(WebAuthnRPError):
  pass


class UnimplementedError(WebAuthnRPError):
  pass


class ValidationError(WebAuthnRPError):
  pass


class NotFoundError(WebAuthnRPError):
  pass


class SignatureCountError(WebAuthnRPError):
  pass


class RegistrationError(WebAuthnRPError):
  pass


class AttestationError(WebAuthnRPError):
  pass



from functools import singledispatch
from typing import Optional, Sequence, Tuple, cast

import cryptography
import cryptography.x509
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric.ec import (
    ECDSA, SECP256R1, EllipticCurvePublicKey)
from cryptography.hazmat.primitives.asymmetric.ed448 import Ed448PublicKey
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PublicKey
from cryptography.hazmat.primitives.hashes import SHA256, SHA384, SHA512
from cryptography.hazmat.primitives.serialization import Encoding, PublicFormat
from cryptography.x509 import Certificate
from cryptography.x509.extensions import UnrecognizedExtension
from cryptography.x509.oid import ObjectIdentifier
from pyasn1.codec.der.decoder import decode
from pyasn1.error import PyAsn1Error

from webauthn_rp.asn1 import KeyDescription
from webauthn_rp.constants import KM_ORIGIN_GENERATED, KM_PURPOSE_SIGN
from webauthn_rp.converters import cryptography_public_key
from webauthn_rp.errors import (AttestationError, UnimplementedError,
                                ValidationError, VerificationError)
from webauthn_rp.types import (AndroidKeyAttestationStatement,
                               AttestationObject, AttestationStatement,
                               AttestationType, CredentialPublicKey,
                               EC2CredentialPublicKey,
                               FIDOU2FAttestationStatement,
                               NoneAttestationStatement, TrustedPath)


@singledispatch
def attest(att_stmt: AttestationStatement, att_obj: AttestationObject,
           auth_data: bytes,
           client_data_hash: bytes) -> Tuple[AttestationType, TrustedPath]:
  raise UnimplementedError('{} verification unimplemented'.format(
      type(att_stmt)))


@attest.register(FIDOU2FAttestationStatement)
def attest_fido_u2f(
    att_stmt: FIDOU2FAttestationStatement, att_obj: AttestationObject,
    auth_data: bytes,
    client_data_hash: bytes) -> Tuple[AttestationType, TrustedPath]:
  """Attest a FIDO U2F key.

  Args:
    att_stmt (FIDOU2FAttestationStatement): The attestation statment.
    att_obj (AttestationObject): The attestation object.
    auth_data (bytes): The raw authenticator data.
    client_data_hash (bytes): The client data hash.

  Returns:
    The attestation type and trusted path.
  
  References:
    * https://www.w3.org/TR/webauthn/#fido-u2f-attestation
  """
  if len(att_stmt.x5c) != 1:
    raise ValidationError(
        'FIDO U2F verification failed: must have a single X.509 certificate')

  att_cert = att_stmt.x5c[0]
  att_cert_x509 = cryptography.x509.load_der_x509_certificate(
      att_cert, default_backend())
  att_cert_x509_pk = att_cert_x509.public_key()
  if not isinstance(att_cert_x509_pk, EllipticCurvePublicKey):
    raise ValidationError(
        'FIDO U2F verification failed: must use an Elliptic Curve Public Key')

  if not isinstance(att_cert_x509_pk.curve, SECP256R1):
    raise ValidationError(
        'FIDO U2F verification failed: must use an Elliptic Curve Public Key')

  assert att_obj.auth_data is not None
  assert att_obj.auth_data.attested_credential_data is not None

  credential_public_key = cast(
      EC2CredentialPublicKey,
      att_obj.auth_data.attested_credential_data.credential_public_key)

  assert credential_public_key is not None
  public_key_u2f = b'\x04' + (credential_public_key.x +
                              credential_public_key.y)

  rp_id_hash = att_obj.auth_data.rp_id_hash
  credential_id = att_obj.auth_data.attested_credential_data.credential_id
  verification_data = (b'\x00' + rp_id_hash +
                       (client_data_hash + credential_id + public_key_u2f))

  assert att_stmt.sig is not None

  try:
    att_cert_x509_pk.verify(att_stmt.sig, verification_data, ECDSA(SHA256()))
  except cryptography.exceptions.InvalidSignature:
    raise VerificationError('FIDO U2F verification failed: invalid signature')

  return AttestationType.BASIC, [att_cert_x509]


@attest.register(AndroidKeyAttestationStatement)
def attest_android_key(
    att_stmt: AndroidKeyAttestationStatement, att_obj: AttestationObject,
    auth_data: bytes,
    client_data_hash: bytes) -> Tuple[AttestationType, TrustedPath]:
  """Attest an android key.

  Args:
    att_stmt (AndroidKeyAttestationStatement): The attestation statment.
    att_obj (AttestationObject): The attestation object.
    auth_data (bytes): The raw authenticator data.
    client_data_hash (bytes): The client data hash.

  Returns:
    The attestation type and trusted path.
  
  References:
    * https://www.w3.org/TR/webauthn/#android-key-attestation
    * https://source.android.com/security/keystore/attestation
    * https://developer.android.com/training/articles/security-key-attestation
  """
  if len(att_stmt.x5c) == 0:
    raise ValidationError('Must have at least 1 X509 certificate')

  credential_certificate = cryptography.x509.load_der_x509_certificate(
      att_stmt.x5c[0], default_backend())
  cred_cert_pk = credential_certificate.public_key()
  if not isinstance(
      cred_cert_pk,
      (EllipticCurvePublicKey, Ed25519PublicKey, Ed448PublicKey)):
    raise ValidationError(
        'Android key verification failed: must use an Elliptic Curve Public Key'
    )

  assert att_stmt.alg is not None
  alg_name = att_stmt.alg.name

  alg_to_hash = {
      'ES256': SHA256,
      'ES384': SHA384,
      'ES512': SHA512,
  }

  hash_algorithm = None
  if alg_name in alg_to_hash: hash_algorithm = ECDSA(alg_to_hash[alg_name]())
  elif alg_name != 'EDDSA':
    raise ValidationError('Unsupported hashing algorithm {}'.format(alg_name))

  assert att_obj.auth_data is not None
  assert att_obj.auth_data.attested_credential_data is not None

  cpk = cryptography_public_key(
      att_obj.auth_data.attested_credential_data.credential_public_key)

  verification_data = auth_data + client_data_hash
  assert att_stmt.sig is not None

  try:
    if hash_algorithm is not None:
      assert isinstance(cred_cert_pk, EllipticCurvePublicKey)
      cred_cert_pk.verify(att_stmt.sig, verification_data, hash_algorithm)
    else:
      assert isinstance(cred_cert_pk, (Ed25519PublicKey, Ed448PublicKey))
      cred_cert_pk.verify(att_stmt.sig, verification_data)
  except cryptography.exceptions.InvalidSignature:
    raise VerificationError(
        'Android Key verification failed: invalid signature')

  cpk_public_bytes = cpk.public_bytes(Encoding.DER,
                                      PublicFormat.SubjectPublicKeyInfo)
  cred_cert_public_bytes = cred_cert_pk.public_bytes(
      Encoding.DER, PublicFormat.SubjectPublicKeyInfo)
  if cpk_public_bytes != cred_cert_public_bytes:
    raise ValidationError(
        ('Certificate public key in attestation statement must match the '
         'provided credential public key'))

  try:
    extension = credential_certificate.extensions.get_extension_for_oid(
        ObjectIdentifier('1.3.6.1.4.1.11129.2.1.17'))
    assert isinstance(extension.value, UnrecognizedExtension)
  except cryptography.x509.ExtensionNotFound:
    raise ValidationError(
        'Could not find android key attestation certificate extension data')

  try:
    key_description, _ = decode(extension.value.value, KeyDescription())
  except PyAsn1Error:
    raise ValidationError(
        'Unable to decode DER-encoded Android Key Description')

  attestation_challenge = key_description['attestationChallenge'].asOctets()
  if attestation_challenge != client_data_hash:
    raise ValidationError(
        'Client data hash does not match value of attestation extension data')

  all_apps_se = key_description['softwareEnforced']['allApplications']
  all_apps_tee = key_description['teeEnforced']['allApplications']
  if all_apps_se.hasValue() or all_apps_tee.hasValue():
    raise ValidationError('The allApplications field must not be present '
                          'in the android key description')

  # TODO: Consider selecting the appropriate AuthorizationList.
  tee_origin = key_description['teeEnforced']['origin']
  tee_purpose = key_description['teeEnforced']['purpose']
  if not tee_origin.hasValue() or int(tee_origin) != KM_ORIGIN_GENERATED:
    raise ValidationError('The teeEnforced origin field must equal '
                          'KM_ORIGIN_GENERATED={}'.format(KM_ORIGIN_GENERATED))

  # TODO: Determine if other purposes are also allowed in this set.
  if not tee_purpose.hasValue() or tee_purpose.count(KM_PURPOSE_SIGN) == 0:
    raise ValidationError('The teeEnforced purpose field must contain '
                          'KM_PURPOSE_SIGN={}'.format(KM_PURPOSE_SIGN))

  return AttestationType.BASIC, [credential_certificate]


@attest.register(NoneAttestationStatement)
def attest_none(
    att_stmt: NoneAttestationStatement, att_obj: AttestationObject,
    auth_data: bytes,
    client_data_hash: bytes) -> Tuple[AttestationType, TrustedPath]:
  return AttestationType.NONE, None



import base64
import json
from enum import Enum
from functools import singledispatch
from typing import Any, Optional, Union

import cbor
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric.ec import (
    SECP256R1, SECP384R1, SECP521R1, EllipticCurvePublicNumbers)
from cryptography.hazmat.primitives.asymmetric.ed448 import Ed448PublicKey
from cryptography.hazmat.primitives.asymmetric.ed25519 import Ed25519PublicKey

from webauthn_rp.errors import UnimplementedError, ValidationError
from webauthn_rp.types import (CredentialPublicKey, EC2CredentialPublicKey,
                               EC2PublicKey, OKPCredentialPublicKey,
                               OKPPublicKey, PublicKey, PublicKeyCredential)
from webauthn_rp.utils import snake_to_camel_case

JSONValue = Union[dict, list, bool, int, float, str]


@singledispatch
def jsonify(data: Any, convert_case: bool = True) -> JSONValue:
  if not isinstance(data, Enum) and hasattr(data, '__dict__'):
    data = data.__dict__

  if isinstance(data, Enum):
    return jsonify(data.value, convert_case)
  elif type(data) is dict:
    for k in data:
      if type(k) is not str:
        raise ValidationError('The type of dict keys must be a string in JSON')

    return {(snake_to_camel_case(k) if convert_case else k):
            jsonify(v, convert_case)
            for k, v in (data.items()) if v is not None}
  elif type(data) is bytes:
    return base64.b64encode(data).decode('utf-8')
  elif type(data) in (str, int, float, bool):
    return data
  elif type(data) in (list, tuple):
    return [jsonify(x, convert_case) for x in data]
  else:
    raise UnimplementedError('JSON conversion for given data is not supported')


@singledispatch
def cryptography_public_key(
    credential_public_key: CredentialPublicKey) -> PublicKey:
  raise UnimplementedError('Must implement public key conversion')


@cryptography_public_key.register(EC2CredentialPublicKey)
def cryptography_ec2_public_key(
    credential_public_key: EC2CredentialPublicKey) -> EC2PublicKey:
  x = int.from_bytes(credential_public_key.x, 'big')
  y = int.from_bytes(credential_public_key.y, 'big')

  curve: Optional[Union[SECP256R1, SECP384R1, SECP521R1]] = None
  if credential_public_key.crv.name == 'P_256': curve = SECP256R1()
  elif credential_public_key.crv.name == 'P_384': curve = SECP384R1()
  elif credential_public_key.crv.name == 'P_521': curve = SECP521R1()
  else:
    raise UnimplementedError('Unsupported cryptography EC2 curve {}'.format(
        credential_public_key.crv.name))

  assert curve is not None

  ecpn = EllipticCurvePublicNumbers(x, y, curve)
  return ecpn.public_key(default_backend())


@cryptography_public_key.register(OKPCredentialPublicKey)
def cryptography_okp_public_key(
    credential_public_key: OKPCredentialPublicKey) -> OKPPublicKey:
  if credential_public_key.crv.name == 'ED25519':
    return Ed25519PublicKey.from_public_bytes(credential_public_key.x)
  elif credential_public_key.crv.name == 'ED448':
    return Ed448PublicKey.from_public_bytes(credential_public_key.x)
  else:
    raise UnimplementedError('Unsupported cryptography OKP curve {}'.format(
        credential_public_key.crv.name))


def build_base_cose_dictionary(
    credential_public_key: CredentialPublicKey) -> dict:
  d = {}
  d[1] = credential_public_key.kty.value
  if credential_public_key.kid is not None:
    d[2] = credential_public_key.kid
  assert credential_public_key.alg is not None
  d[3] = credential_public_key.alg.value
  if credential_public_key.key_ops is not None:
    d[4] = [x.value for x in credential_public_key.key_ops]
  if credential_public_key.key_ops is not None:
    d[5] = credential_public_key.base_IV
  return d


@singledispatch
def cose_key(credential_public_key: CredentialPublicKey) -> bytes:
  raise UnimplementedError('Must implement cose key conversion')


@cose_key.register(EC2CredentialPublicKey)
def cose_key_from_ec2(credential_public_key: EC2CredentialPublicKey) -> bytes:
  d = build_base_cose_dictionary(credential_public_key)
  d[-1] = credential_public_key.crv.value
  d[-2] = credential_public_key.x
  d[-3] = credential_public_key.y
  return cbor.dumps(d)


@cose_key.register(OKPCredentialPublicKey)
def cose_key_from_okp(credential_public_key: OKPCredentialPublicKey) -> bytes:
  d = build_base_cose_dictionary(credential_public_key)
  d[-1] = credential_public_key.crv.value
  d[-2] = credential_public_key.x
  return cbor.dumps(d)



from typing import Any, NamedTuple, Optional, Sequence, Union

from webauthn_rp.errors import UnimplementedError
from webauthn_rp.types import (
    AttestationObject, AttestationStatementFormatIdentifier, AttestationType,
    AuthenticatorAssertionResponse, AuthenticatorAttestationResponse,
    AuthenticatorData, CredentialCreationOptions, CredentialPublicKey,
    CredentialRequestOptions, PublicKey, PublicKeyCredential,
    PublicKeyCredentialRpEntity, PublicKeyCredentialUserEntity, TrustedPath)


class CredentialData(NamedTuple):
  credential_public_key: CredentialPublicKey
  signature_count: int
  user_entity: Optional[PublicKeyCredentialUserEntity] = None
  rp_entity: Optional[PublicKeyCredentialRpEntity] = None


class CredentialsRegistrar:
  def register_creation_options(self,
                                options: CredentialCreationOptions) -> bool:
    return True

  def register_request_options(self,
                               options: CredentialRequestOptions) -> bool:
    return True

  def register_credential_creation(
      self,
      credential: PublicKeyCredential,
      att: AttestationObject,
      att_type: AttestationType,
      user: PublicKeyCredentialUserEntity,
      rp: PublicKeyCredentialRpEntity,
      trusted_path: Optional[TrustedPath] = None) -> bool:
    raise UnimplementedError('Must implement register_credential_creation')

  def register_credential_request(self, credential: PublicKeyCredential,
                                  authenticator_data: AuthenticatorData,
                                  user: PublicKeyCredentialUserEntity,
                                  rp: PublicKeyCredentialRpEntity) -> bool:
    raise UnimplementedError('Must implement register_credential_request')

  def get_credential_data(self,
                          credential_id: bytes) -> Optional[CredentialData]:
    raise UnimplementedError('Must implement get_credential_data')

  def check_user_owns_credential(self, user_handle: bytes,
                                 credential_id: bytes) -> Optional[bool]:
    raise UnimplementedError('Must implement check_user_owns_credential')



from enum import Enum
from typing import Any, ClassVar, Optional, Sequence, Type, TypeVar, Union

from cryptography.hazmat.primitives.asymmetric.ec import (
    EllipticCurvePrivateKey, EllipticCurvePublicKey)
from cryptography.hazmat.primitives.asymmetric.ed448 import (Ed448PrivateKey,
                                                             Ed448PublicKey)
from cryptography.hazmat.primitives.asymmetric.ed25519 import (
    Ed25519PrivateKey, Ed25519PublicKey)
from cryptography.hazmat.primitives.asymmetric.rsa import RSAPublicKey
from cryptography.x509 import Certificate

from webauthn_rp.utils import camel_to_snake_case

EC2PublicKey = Union[EllipticCurvePublicKey]
OKPPublicKey = Union[Ed25519PublicKey, Ed448PublicKey]

EC2PrivateKey = Union[EllipticCurvePrivateKey]
OKPPrivateKey = Union[Ed25519PrivateKey, Ed448PrivateKey]

PublicKey = Union[EC2PublicKey, OKPPublicKey]
PrivateKey = Union[EC2PrivateKey, OKPPrivateKey]
TrustedPath = Optional[Sequence[Certificate]]


class PublicKeyCredentialEntity:
  """
  The PublicKeyCredentialEntity describes a user account, or a WebAuthn Relying
  Party, which a public key credential is associated with or scoped to,
  respectively.
  
  Attributes:
    name (str): A human-palatable name for the entity.
      Its function depends on what the PublicKeyCredentialEntity represents:

      * When inherited by PublicKeyCredentialRpEntity it is a human-palatable
        identifier for the Relying Party, intended only for display.
      * When inherited by PublicKeyCredentialUserEntity, it is a
        human-palatable identifier for a user account. It is intended only
        for display, i.e., aiding the user in determining the difference
        between user accounts with similar displayNames.
    icon (str):
      A serialized URL which resolves to an image associated with the entity.
      For example, this could be a user’s avatar or a Relying Party's logo.
      This URL MUST be an a priori authenticated URL. Authenticators MUST
      accept and store a 128-byte minimum length for an icon member’s value.
      Authenticators MAY ignore an icon member’s value if its length is
      greater than 128 bytes. The URL’s scheme MAY be "data" to avoid fetches
      of the URL, at the cost of needing more storage.

  References:
    * https://w3.org/TR/webauthn/#dictdef-publickeycredentialentity
  """
  def __init__(self, *, name: str, icon: Optional[str] = None):
    self.name = name
    self.icon = icon


class PublicKeyCredentialRpEntity(PublicKeyCredentialEntity):
  """
  The PublicKeyCredentialRpEntity is used to supply additional
  Relying Party attributes when creating a new credential.

  Attributes:
    name (str): A human-palatable name for the entity.
      Its function depends on what the PublicKeyCredentialEntity represents:

      * When inherited by PublicKeyCredentialRpEntity it is a human-palatable
        identifier for the Relying Party, intended only for display.
      * When inherited by PublicKeyCredentialUserEntity, it is a
        human-palatable identifier for a user account. It is intended only
        for display, i.e., aiding the user in determining the difference
        between user accounts with similar displayNames.
    icon (str):
      A serialized URL which resolves to an image associated with the entity.
      For example, this could be a user’s avatar or a Relying Party's logo.
      This URL MUST be an a priori authenticated URL. Authenticators MUST
      accept and store a 128-byte minimum length for an icon member’s value.
      Authenticators MAY ignore an icon member’s value if its length is
      greater than 128 bytes. The URL’s scheme MAY be "data" to avoid fetches
      of the URL, at the cost of needing more storage.
    id (str): A unique identifier for the Relying Party entity.

  References:
    * https://w3.org/TR/webauthn/#dictdef-publickeycredentialrpentity
    * https://w3.org/TR/webauthn/#dictdef-publickeycredentialentity
  """
  def __init__(self, *, name: str, icon: Optional[str] = None, id: str):
    super().__init__(name=name, icon=icon)
    self.id = id


class PublicKeyCredentialUserEntity(PublicKeyCredentialEntity):
  """
  The PublicKeyCredentialUserEntity is used to supply additional user account
  attributes when creating a new credential.

  Attributes:
    name (str): A human-palatable name for the entity.
      Its function depends on what the PublicKeyCredentialEntity represents:

      * When inherited by PublicKeyCredentialRpEntity it is a human-palatable
        identifier for the Relying Party, intended only for display.
      * When inherited by PublicKeyCredentialUserEntity, it is a
        human-palatable identifier for a user account. It is intended only
        for display, i.e., aiding the user in determining the difference
        between user accounts with similar displayNames.
    icon (str):
      A serialized URL which resolves to an image associated with the entity.
      For example, this could be a user’s avatar or a Relying Party's logo.
      This URL MUST be an a priori authenticated URL. Authenticators MUST
      accept and store a 128-byte minimum length for an icon member’s value.
      Authenticators MAY ignore an icon member’s value if its length is
      greater than 128 bytes. The URL’s scheme MAY be "data" to avoid fetches
      of the URL, at the cost of needing more storage.
    id (bytes): The user handle of the user account entity. 
      To ensure secure operation, authentication and authorization decisions
      MUST be made on the basis of this id member, not the displayName nor
      name members.

      Since the user handle (id) is not considered personally identifying
      information, the Relying Party SHOULD NOT include personally identifying
      information, e.g., e-mail addresses or usernames, in the user handle.
      This includes hash values of personally identifying information, unless
      the hash function is salted with salt values private to the Relying
      Party, since hashing does not prevent probing for guessable input values.
      It is RECOMMENDED to let the user handle be 64 random bytes, and store
      this value in the user’s account.
    display_name (str): A human-palatable name for the user account, intended
      only for display. The Relying Party SHOULD let the user choose this,
      and SHOULD NOT restrict the choice more than necessary.
  
  References:
    * https://w3.org/TR/webauthn/#dictdef-publickeycredentialuserentity
    * https://w3.org/TR/webauthn/#dictdef-publickeycredentialentity
    * https://w3.org/TR/webauthn/#user-handle
    * https://w3.org/TR/webauthn/#sctn-user-handle-privacy
  """
  def __init__(self,
               *,
               name: str,
               icon: Optional[str] = None,
               id: bytes,
               display_name: str):
    super().__init__(name=name, icon=icon)
    self.id = id
    self.display_name = display_name


class PublicKeyCredentialType(Enum):
  """
  This enumeration defines the valid credential types. It is an extension
  point; values can be added to it in the future, as more credential types are
  defined. The values of this enumeration are used for versioning the
  Authentication Assertion and attestation structures according to the type of
  the authenticator.

  Attributes:
    PUBLIC_KEY (str): The only credential type.

  References:
    * https://w3.org/TR/webauthn/#enumdef-publickeycredentialtype
  """

  PUBLIC_KEY = 'public-key'


class AuthenticatorTransport(Enum):
  """
  Authenticators may implement various transports for communicating with
  clients. This enumeration defines hints as to how clients might communicate
  with a particular authenticator in order to obtain an assertion for a
  specific credential. Note that these hints represent the WebAuthn Relying
  Party's best belief as to how an authenticator may be reached. A Relying
  Party may obtain a list of transports hints from some attestation statement
  formats or via some out-of-band mechanism; it is outside the scope of this
  specification to define that mechanism.

  Attributes:
    USB (str): Indicates the respective authenticator can be contacted over
      removable USB.
    NFC (str): Indicates the respective authenticator can be contacted over
      Near Field Communication (NFC).
    BLE (str): Indicates the respective authenticator can be contacted over
      Bluetooth Smart (Bluetooth Low Energy / BLE).
    INTERNAL (str): Indicates the respective authenticator is contacted using
      a client device-specific transport. These authenticators are not
      removable from the client device.

  References:
    * https://w3.org/TR/webauthn/#enumdef-authenticatortransport
  """

  USB = 'usb'
  NFC = 'nfc'
  BLE = 'ble'
  INTERNAL = 'internal'


class NameValueEnumsContainer(type):
  """
  A metaclass used to enable Enums that can take on a string and integer type
  for a name and value respectively.
  """
  def __call__(cls: Any, value: Union[int, str]) -> Any:  # type: ignore
    if type(value) is int:
      return cls.Value(value)
    elif type(value) is str:
      return cls.Name(value)
    else:
      raise KeyError('Invalid key {}'.format(value))

  def __getitem__(cls: Any, value: Union[int, str]) -> Any:  # type: ignore
    if type(value) is int:
      return cls.Value[value]
    elif type(value) is str:
      return cls.Name[value]
    else:
      raise KeyError('Invalid key {}'.format(value))


class COSEAlgorithmIdentifier(metaclass=NameValueEnumsContainer):
  """
  A COSEAlgorithmIdentifier's value is a number identifying a cryptographic
  algorithm. The algorithm identifiers SHOULD be values registered in the
  IANA COSE Algorithms registry.

  This Enum only contains algorithms that are internally supported. It can be
  extended upon further support. As Web Authentication mainly performs signing
  and verification, only algorithms that pertain to those operations are
  considered.

  References:
    * https://w3.org/TR/webauthn/#typedefdef-cosealgorithmidentifier
    * https://iana.org/assignments/cose/cose.xhtml#algorithms
  """
  class Name(Enum):
    ES256 = 'ES256'
    ES384 = 'ES384'
    ES512 = 'ES512'

    EDDSA = 'EdDSA'

  class Value(Enum):
    ES256 = -7
    ES384 = -35
    ES512 = -36

    EDDSA = -8


class PublicKeyCredentialParameters:
  """
  PublicKeyCredentialParameters is used to supply additional parameters when
  creating a new credential.

  Attributes:
    type (PublicKeyCredentialType):
      This member specifies the type of credential to be created.
    alg (COSEAlgorithmIdentifier):
      This member specifies the cryptographic signature algorithm with which
      the newly generated credential will be used, and thus also the type of
      asymmetric key pair to be generated, e.g., RSA or Elliptic Curve.
  
  References:
    * https://w3.org/TR/webauthn/#dictdef-publickeycredentialparameters
  """
  def __init__(self, *, type: PublicKeyCredentialType,
               alg: Union[COSEAlgorithmIdentifier.Name,
                          COSEAlgorithmIdentifier.Value]):
    self.type = type
    self.alg = alg


class PublicKeyCredentialDescriptor:
  """
  The PublicKeyCredentialDescriptor contains the attributes that are specified
  by a caller when referring to a public key credential as an input parameter
  to the navigator.credentials.create() or navigator.credentials.get() methods
  (on the client side). It mirrors the fields of the PublicKeyCredential object
  returned by the latter methods.

  Attributes:
    type (PublicKeyCredentialType):
      This member contains the type of the public key credential the caller
      is referring to.
    id (bytes):
      This member contains the credential ID of the public key credential
      the caller is referring to.
    transports (Sequence[AuthenticatorTransport]):
      This OPTIONAL member contains a hint as to how the client might
      communicate with the managing authenticator of the public key
      credential the caller is referring to.
  
  References:
    * https://w3.org/TR/webauthn/#dictdef-publickeycredentialdescriptor
  """
  def __init__(self,
               *,
               type: PublicKeyCredentialType,
               id: bytes,
               transports: Optional[Sequence[AuthenticatorTransport]] = None):
    self.type = type
    self.id = id
    self.transports = transports


class AuthenticatorAttachment(Enum):
  """
  This enumeration’s values describe authenticators' attachment modalities.
  Relying Parties use this for two purposes:
  
    * to express a preferred authenticator attachment modality when calling
      navigator.credentials.create() to create a credential, and
    * to inform the client of the Relying Party's best belief about how to
      locate the managing authenticators of the credentials listed in
      allowCredentials when calling navigator.credentials.get() (on the
      client side).

  Attributes:
    PLATFORM (str):
      This value indicates platform attachment.
    CROSS_PLATFORM (str):
      This value indicates cross-platform attachment.

  References:
    * https://w3.org/TR/webauthn/#enumdef-authenticatorattachment
  """

  PLATFORM = 'platform'
  CROSS_PLATFORM = 'cross-platform'


class AttestationConveyancePreference(Enum):
  """
  WebAuthn Relying Parties may use AttestationConveyancePreference to specify
  their preference regarding attestation conveyance during credential
  generation.

  Attributes:
    NONE (str):
      This value indicates that the Relying Party is not interested
      in authenticator attestation. For example, in order to potentially avoid
      having to obtain user consent to relay identifying information to the
      Relying Party, or to save a roundtrip to an Attestation CA.
    INDIRECT (str):
      This value indicates that the Relying Party prefers an
      attestation conveyance yielding verifiable attestation statements, but
      allows the client to decide how to obtain such attestation statements.
      The client MAY replace the authenticator-generated attestation statements
      with attestation statements generated by an Anonymization CA, in order to
      protect the user’s privacy, or to assist Relying Parties with attestation
      verification in a heterogeneous ecosystem.
    DIRECT (str):
      This value indicates that the Relying Party wants to receive
      the attestation statement as generated by the authenticator.

  References:
    * https://w3.org/TR/webauthn/#enumdef-attestationconveyancepreference
  """

  NONE = 'none'
  INDIRECT = 'indirect'
  DIRECT = 'direct'


class UserVerificationRequirement(Enum):
  """
  A WebAuthn Relying Party may require user verification for some of its
  operations but not for others, and may use this type to express its needs.

  Attributes:
    REQUIRED (str):
      This value indicates that the Relying Party requires user verification
      for the operation and will fail the operation if the response does not
      have the UV flag set.
    PREFERRED (str):
      This value indicates that the Relying Party prefers user verification
      for the operation if possible, but will not fail the operation if the
      response does not have the UV flag set.
    DISCOURAGED (str):
      This value indicates that the Relying Party does not want user
      verification employed during the operation (e.g., in the interest of
      minimizing disruption to the user interaction flow).

  References:
    * https://w3.org/TR/webauthn/#enumdef-userverificationrequirement
  """

  REQUIRED = 'required'
  PREFERRED = 'preferred'
  DISCOURAGED = 'discouraged'


class TxAuthGenericArg:
  """
  The input type of the txAuthGeneric extension.

  Attributes:
    content_type (str): MIME-Type of the content, e.g., "image/png"
    content (bytes): The byte string data.

  References
    * https://w3.org/TR/webauthn/#sctn-generic-txauth-extension
  """
  def __init__(self, *, content_type: str, content: bytes):
    self.content_type = content_type
    self.content = content


class Coordinates:
  """
  The geographic coordinate reference system used by the attributes in this
  interface is the World Geodetic System (2d) [WGS84]. No other reference
  system is supported.

  Attributes:
    latitude (float): Geographic latitude specified in decimal degrees.
    longitude (float): Geographic longitude specified in decimal degrees.
    altitude (Optional[float]): Denotes the height of the position, specified
      in meters above the ellipsoid established by
      http://earth-info.nga.mil/GandG/publications/tr8350.2/wgs84fin.pdf.
      If the implementation cannot provide altitude information, the value of
      this attribute must be null.
    accuracy (float):
      Denotes the accuracy level of the latitude and longitude coordinates.
      It is specified in meters and must be supported by all implementations.
      The value of the accuracy attribute must be a non-negative real number.
    altitude_accuracy (Optional[float]):
      Specified in meters. If the implementation cannot provide altitude
      information, the value of this attribute must be null. Otherwise, the
      value must be a non-negative real number.
    heading (Optional[float]):
      Denotes the direction of travel of the hosting device and is specified in
      degrees, where 0° ≤ heading < 360°, counting clockwise relative to the
      true north. If the implementation cannot provide heading information, the
      value of this attribute must be null. If the hosting device is stationary
      (i.e. the value of the speed attribute is 0), then the value of the
      heading attribute must be NaN.
    speed (Optional[float]):
      Denotes the magnitude of the horizontal component of the hosting device's
      current velocity and is specified in meters per second. If the
      implementation cannot provide speed information, the value of this
      attribute must be null. Otherwise, the value of the speed attribute must
      be a non-negative real number.
    
    Notes:
      * The accuracy and altitude_accuracy values returned by an implementation
        should correspond to a 95% confidence level.

    References:
      * https://w3.org/TR/webauthn/#sctn-location-extension
      * https://w3.org/TR/geolocation-API/#coordinates_interface
  """
  def __init__(self,
               *,
               latitude: float,
               longitude: float,
               altitude: Optional[float] = None,
               accuracy: float,
               altitude_accuracy: Optional[float] = None,
               heading: Optional[float] = None,
               speed: Optional[float] = None):
    self.latitude = latitude
    self.longitude = longitude
    self.altitude = altitude
    self.accuracy = accuracy
    self.altitude_accuracy = altitude_accuracy
    self.heading = heading
    self.speed = speed


AAGUID = bytes
AuthenticatorSelectionList = Sequence[AAGUID]
AuthenticationExtensionsSupported = Sequence[str]
UvmEntry = Sequence[int]
UvmEntries = Sequence[UvmEntry]


class AuthenticatorBiometricPerfBounds:
  """
  Biometric performance bounds.

  Attributes:
    FAR (float): The maximum false acceptance rate for a biometric
      authenticator allowed by the Relying Party.
    FRR (float): The maximum false rejection rate for a biometric authenticator
      allowed by the Relying Party.
  
  References:
    * w3.org/TR/webauthn/#sctn-authenticator-biometric-criteria-extension
  """
  def __init__(self, *, FAR: float, FRR: float):
    self.FAR = FAR
    self.FRR = FRR


class ExtensionIdentifier(Enum):
  """
  The string identifier of supported extensions.

  Attributes:
    APPID (str):
      This extension allows WebAuthn Relying Parties that have previously
      registered a credential using the legacy FIDO JavaScript APIs to request
      an assertion. The FIDO APIs use an alternative identifier for Relying
      Parties called an AppID [FIDO-APPID], and any credentials created using
      those APIs will be scoped to that identifier. Without this extension,
      they would need to be re-registered in order to be scoped to an RP ID.

      This extension does not allow FIDO-compatible credentials to be created.
      Thus, credentials created with WebAuthn are not backwards compatible with
      the FIDO JavaScript APIs.
    TX_AUTH_SIMPLE (str):
      This extension allows for a simple form of transaction authorization.
      A Relying Party can specify a prompt string, intended for display on a
      trusted device on the authenticator.
    TX_AUTH_GENERIC (str):
      This extension allows images to be used as transaction authorization
      prompts as well. This allows authenticators without a font rendering
      engine to be used and also supports a richer visual appearance.
    AUTHN_SEL (str):
      This extension allows a WebAuthn Relying Party to guide the selection of
      the authenticator that will be leveraged when creating the credential.
      It is intended primarily for Relying Parties that wish to tightly control
      the experience around credential creation.
    EXTS (str):
      This extension enables the WebAuthn Relying Party to determine which
      extensions the authenticator supports.
    UVI (str):
      This extension enables use of a user verification index.
    LOC (str):
      This extension provides the authenticator's current location to the
      WebAuthn Relying Party.
    UVM (str);
      This extension enables use of a user verification method.
    BIOMETRIC_PERF_BOUNDS (str):
      This extension allows WebAuthn Relying Parties to specify the desired
      performance bounds for selecting biometric authenticators as candidates
      to be employed in a registration ceremony.

  References:
    * https://w3.org/TR/webauthn/#sctn-defined-extensions
  """

  APPID = 'appid'
  TX_AUTH_SIMPLE = 'txAuthSimple'
  TX_AUTH_GENERIC = 'txAuthGeneric'
  AUTHN_SEL = 'authnSel'
  EXTS = 'exts'
  UVI = 'uvi'
  LOC = 'loc'
  UVM = 'uvm'
  BIOMETRIC_PERF_BOUNDS = 'biometricPerfBounds'

  @property
  def key(self):
    return camel_to_snake_case(self.value)


class AuthenticationExtensionsClientInputs:
  """
  This is an object containing the client extension input values for zero or
  more WebAuthn extensions.

  Attributes:
    appid (Optional[str]): A single string specifying a FIDO AppID.
    tx_auth_simple (Optional[str]): A single string prompt.
    tx_auth_generic (Optional[TxAuthGenericArg]):
      An object containing generic inputs.
    authn_sel (Optional[AuthenticatorSelectionList]):
      A sequence of AAGUIDs. Each AAGUID corresponds to an authenticator model
      that is acceptable to the Relying Party for this credential creation. The
      list is ordered by decreasing preference.

      An AAGUID is defined as an array containing the globally unique
      identifier of the authenticator model being sought.
    exts (Optional[bool]):
      The Boolean value true to indicate that this extension is requested by
      the Relying Party.
    uvi (Optional[bool]):
      The Boolean value true to indicate that this extension is requested by
      the Relying Party.
    loc (Optional[bool]):
      The Boolean value true to indicate that this extension is requested by
      the Relying Party.
    uvm (Optional[bool]):
      The Boolean value true to indicate that this extension is requested by
      the Relying Party.
    biometric_perf_bounds (Optional[AuthenticatorBiometricPerfBounds]):
      The biometric performance bounds.

  References:
    * https://w3.org/TR/webauthn/#extensions
    * https://w3.org/TR/webauthn/#dictdef-authenticationextensionsclientinputs
    * https://w3.org/TR/webauthn/#dictdef-authenticationextensionsclientoutputs
  """
  def __init__(
      self,
      *,
      appid: Optional[str] = None,
      tx_auth_simple: Optional[str] = None,
      tx_auth_generic: Optional[TxAuthGenericArg] = None,
      authn_sel: Optional[AuthenticatorSelectionList] = None,
      exts: Optional[bool] = None,
      uvi: Optional[bool] = None,
      loc: Optional[bool] = None,
      uvm: Optional[bool] = None,
      biometric_perf_bounds: Optional[AuthenticatorBiometricPerfBounds] = None
  ):
    self.appid = appid
    self.tx_auth_simple = tx_auth_simple
    self.tx_auth_generic = tx_auth_generic
    self.authn_sel = authn_sel
    self.exts = exts
    self.uvi = uvi
    self.loc = loc
    self.uvm = uvm
    self.biometric_perf_bounds = biometric_perf_bounds


class AuthenticationExtensionsClientOutputs:
  """
  This is an object containing the client extension output values for zero or
  more WebAuthn extensions.

  Attributes:
    appid (Optional[bool]): If true, the AppID was used and thus, when
      verifying an assertion, the Relying Party MUST expect the rpIdHash to be
      the hash of the AppID, not the RP ID.
    tx_auth_simple (Optional[str]):
      The authenticator extension output string UTF-8 decoded into a string.
    tx_auth_generic (Optional[bytes]): A generic byte string.
    authn_sel (Optional[AuthenticatorSelectionList]):
      Returns the value true to indicate to the Relying Party that the
      extension was acted upon.
    exts (Optional[AuthenticationExtensionsSupported]):
      Returns the list of supported extensions as an array of extension
      identifier strings.
    uvi (Optional[bytes]):
      Returns the authenticator extension output as a byte string.
    loc (Optional[Coordinates]):
      Returns an object that encodes the location information in the
      authenticator extension output as a Coordinates value.
    uvm (Optional[UvmEntries]):
      Returns a sequence of UvmEntry objects which are 3-element arrays of
      numbers that encode the factors in the authenticator extension output.
    biometric_perf_bounds (Optional[bool]):
      Returns the value true to indicate to the Relying Party that the
      extension was acted upon.

  References:
    * https://w3.org/TR/webauthn/#dictdef-authenticationextensionsclientoutputs
  """
  def __init__(self,
               *,
               appid: Optional[bool] = None,
               tx_auth_simple: Optional[str] = None,
               tx_auth_generic: Optional[bytes] = None,
               authn_sel: Optional[bool] = None,
               exts: Optional[AuthenticationExtensionsSupported] = None,
               uvi: Optional[bytes] = None,
               loc: Optional[Coordinates] = None,
               uvm: Optional[UvmEntries] = None,
               biometric_perf_bounds: Optional[bool] = None):
    self.appid = appid
    self.tx_auth_simple = tx_auth_simple
    self.tx_auth_generic = tx_auth_generic
    self.authn_sel = authn_sel
    self.exts = exts
    self.uvi = uvi
    self.loc = loc
    self.uvm = uvm
    self.biometric_perf_bounds = biometric_perf_bounds


class AuthenticatorSelectionCriteria:
  """
  WebAuthn Relying Parties may use the AuthenticatorSelectionCriteria to
  specify their requirements regarding authenticator attributes.

  Attributes:
    authenticator_attachment (AuthenticatorAttachment):
      If this member is present, eligible authenticators are filtered to only
      authenticators attached with the specified Authenticator Attachment
      Enumeration (enum AuthenticatorAttachment).
    require_resident_key (bool):
      This member describes the Relying Party's requirements regarding
      resident credentials. If the parameter is set to true, the
      authenticator MUST create a client-side-resident public key credential
      source when creating a public key credential.
    user_verification (UserVerificationRequirement):
      This member describes the Relying Party's requirements regarding user
      verification for the navigator.credentials.create() operation
      (on the client side). Eligible authenticators are filtered to only those
      capable of satisfying this requirement.

  References:
    * https://w3.org/TR/webauthn/#dictdef-authenticatorselectioncriteria
  """
  def __init__(
      self,
      *,
      authenticator_attachment: Optional[AuthenticatorAttachment] = None,
      require_resident_key: bool = False,
      user_verification:
      UserVerificationRequirement = UserVerificationRequirement.PREFERRED):
    self.authenticator_attachment = authenticator_attachment
    self.require_resident_key = require_resident_key
    self.user_verification = user_verification


class PublicKeyCredentialCreationOptions:
  """Options for Credential Creation

  Attributes:
    rp (PublicKeyCredentialRpEntity): This member contains data about the
      Relying Party responsible for the request.
    user (PublicKeyCredentialUserEntity): This member contains data about the
      user account for which the Relying Party is requesting attestation.
    challenge (bytes): This member contains a challenge intended to be used
      for generating the newly created credential’s attestation object.
    pub_key_cred_params (Sequence[PublicKeyCredentialParameters]):
      This member contains information about the desired properties of the
      credential to be created. The sequence is ordered from most preferred
      to least preferred. The client makes a best-effort to create the most
      preferred credential that it can.
    timeout (int): This member specifies a time, in milliseconds, that the
      caller is willing to wait for the call to complete. This is treated as
      a hint, and MAY be overridden by the client.
    authenticator_selection (AuthenticatorSelectionCriteria):
      This member is intended for use by Relying Parties that wish to select
      the appropriate authenticators to participate in the
      navigator.credentials.create() operation (on the client side).
    extensions (AuthenticationExtensionsClientInputs):
      This member contains additional parameters requesting additional
      processing by the client and authenticator. For example, the caller may
      request that only authenticators with certain capabilities be used to
      create the credential, or that particular information be returned in
      the attestation object.
    attestation (AttestationConveyancePreference):
      This member is intended for use by Relying Parties that wish to express
      their preference for attestation conveyance.
    exclude_credentials (Sequence[PublicKeyCredentialDescriptor]):
      This member is intended for use by Relying Parties that wish to limit
      the creation of multiple credentials for the same account on a single
      authenticator. The client is requested to return an error if the new
      credential would be created on an authenticator that also contains one
      of the credentials enumerated in this parameter.

  References:
    * https://w3.org/TR/webauthn/#dictdef-publickeycredentialcreationoptions
  """
  def __init__(
      self,
      *,
      rp: PublicKeyCredentialRpEntity,
      user: PublicKeyCredentialUserEntity,
      challenge: bytes,
      pub_key_cred_params: Sequence[PublicKeyCredentialParameters],
      timeout: Optional[int] = None,
      authenticator_selection: Optional[AuthenticatorSelectionCriteria] = None,
      extensions: Optional[AuthenticationExtensionsClientInputs] = None,
      attestation:
      AttestationConveyancePreference = AttestationConveyancePreference.NONE,
      exclude_credentials: Optional[
          Sequence[PublicKeyCredentialDescriptor]] = None):
    self.rp = rp
    self.user = user
    self.challenge = challenge
    self.pub_key_cred_params = pub_key_cred_params
    self.timeout = timeout
    self.exclude_credentials = exclude_credentials
    self.authenticator_selection = authenticator_selection
    self.attestation = attestation
    self.extensions = extensions


class PublicKeyCredentialRequestOptions:
  """
  The PublicKeyCredentialRequestOptions object supplies
  navigator.credentials.get() with the data it needs to generate an assertion
  (on the client side). Its challenge member MUST be present, while its other
  members are OPTIONAL.

  Attributes:
    challenge (bytes): This member represents a challenge that the selected
      authenticator signs, along with other data, when producing an
      authentication assertion.
    timeout (int): This OPTIONAL member specifies a time, in milliseconds,
      that the caller is willing to wait for the call to complete. The value
      is treated as a hint, and MAY be overridden by the client.
    rp_id (str): This OPTIONAL member specifies the relying party identifier
      claimed by the caller. If omitted, its value will be the
      CredentialsContainer object’s relevant settings object's origin's
      effective domain.
    extensions (AuthenticationExtensionsClientInputs): 
    allow_credentials (Sequence[PublicKeyCredentialDescriptor]):
      This OPTIONAL member contains a list of PublicKeyCredentialDescriptor
      objects representing public key credentials acceptable to the caller,
      in descending order of the caller’s preference (the first item in the
      list is the most preferred credential, and so on down the list).
    user_verification (UserVerificationRequirement):
      This OPTIONAL member describes the Relying Party's requirements
      regarding user verification for the navigator.credentials.get() operation
      (on the client side). Eligible authenticators are filtered to only those
      capable of satisfying this requirement.

  References:
    * https://w3.org/TR/webauthn/#dictdef-publickeycredentialrequestoptions
  """
  def __init__(
      self,
      *,
      challenge: bytes,
      timeout: Optional[int] = None,
      rp_id: Optional[str] = None,
      extensions: Optional[AuthenticationExtensionsClientInputs] = None,
      allow_credentials: Optional[
          Sequence[PublicKeyCredentialDescriptor]] = None,
      user_verification: Optional[UserVerificationRequirement] = (
          UserVerificationRequirement.PREFERRED)):
    self.challenge = challenge
    self.timeout = timeout
    self.rp_id = rp_id
    self.allow_credentials = allow_credentials
    self.user_verification = user_verification
    self.extensions = extensions


class CredentialCreationOptions:
  """
  The object used for registration of credentials using the
  navigator.credentials.create() function on the client side.

  Attributes:
    public_key (PublicKeyCredentialCreationOptions):
      The creation options for the public key credential.

  References:
    * https://w3.org/TR/webauthn/#credentialcreationoptions-extension
  """
  def __init__(self, *, public_key: PublicKeyCredentialCreationOptions):
    self.public_key = public_key


class CredentialMediationRequirement(Enum):
  """
  When making a request via get(options), developers can set a case-by-case
  requirement for user mediation by choosing the appropriate
  CredentialMediationRequirement enum value.

  Attributes:
    SILENT (str):
      User mediation is suppressed for the given operation. If the operation
      can be performed without user involvement, wonderful. If user involvement
      is necessary, then the operation will return null rather than involving
      the user.
    OPTIONAL (str):
      If credentials can be handed over for a given operation without user
      mediation, they will be. If user mediation is required, then the user
      agent will involve the user in the decision.
    REQUIRED (str):
      The user agent will not hand over credentials without user mediation,
      even if the prevent silent access flag is unset for an origin.
  
  References:
    * https://w3.org/TR/credential-management-1/#mediation-requirements
  """

  SILENT = 'silent'
  OPTIONAL = 'optional'
  REQUIRED = 'required'


class CredentialRequestOptions:
  """
  The object used to obtain assertions using the
  navigator.credentials.get() function on the client side.

  Attributes:
    mediation (CredentialMediationRequirement):
      Specifies the mediation requirements for a given credential request.
    public_key (PublicKeyCredentialRequestOptions):
      The request options for the public key credential.

  References:
    * https://w3.org/TR/webauthn/#credentialrequestoptions-extension
    * w3.org/TR/credential-management-1/#dictdef-credentialrequestoptions
  """
  def __init__(self,
               *,
               mediation: CredentialMediationRequirement = (
                   CredentialMediationRequirement.OPTIONAL),
               public_key: PublicKeyCredentialRequestOptions):
    self.mediation = mediation
    self.public_key = public_key


class AuthenticatorResponse:
  """
  Authenticators respond to Relying Party requests by returning an object
  derived from an AuthenticatorResponse.

  Attributes:
    client_data_JSON (bytes): This attribute contains a JSON serialization of
      the client data passed to the authenticator by the client in its call
      to either navigator.credentials.create() or navigator.credentials.get().

  References:
    * https://w3.org/TR/webauthn/#authenticatorresponse
  """
  def __init__(self, *, client_data_JSON: bytes):
    self.client_data_JSON = client_data_JSON


class AuthenticatorAttestationResponse(AuthenticatorResponse):
  """
  The AuthenticatorAttestationResponse represents the authenticator's response
  to a client’s request for the creation of a new public key credential. It
  contains information about the new credential that can be used to identify
  it for later use, and metadata that can be used by the WebAuthn Relying Party
  to assess the characteristics of the credential during registration.

  Attributes:
    client_data_JSON (bytes): This attribute contains a JSON serialization of
      the client data passed to the authenticator by the client in its call
      to either navigator.credentials.create() or navigator.credentials.get().
    attestation_object (bytes): This attribute contains an attestation object,
      which is opaque to, and cryptographically protected against tampering by,
      the client. The attestation object contains both authenticator data and
      an attestation statement. The former contains the AAGUID, a unique
      credential ID, and the credential public key. The contents of the
      attestation statement are determined by the attestation statement
      format used by the authenticator. It also contains any additional
      information that the Relying Party's server requires to validate the
      attestation statement, as well as to decode and validate the
      authenticator data along with the JSON-serialized client data.

  References:
    * https://w3.org/TR/webauthn/#authenticatorattestationresponse
    * https://w3.org/TR/webauthn/#authenticatorresponse
  """
  def __init__(self, *, client_data_JSON: bytes, attestation_object: bytes):
    super().__init__(client_data_JSON=client_data_JSON)
    self.attestation_object = attestation_object


class AuthenticatorAssertionResponse(AuthenticatorResponse):
  """
  The AuthenticatorAssertionResponse interface represents an authenticator's
  response to a client’s request for generation of a new authentication
  assertion given the WebAuthn Relying Party's challenge and OPTIONAL list
  of credentials it is aware of. This response contains a cryptographic
  signature proving possession of the credential private key, and optionally
  evidence of user consent to a specific transaction.

  Attributes:
    client_data_JSON (bytes): This attribute contains a JSON serialization of
      the client data passed to the authenticator by the client in its call
      to either navigator.credentials.create() or navigator.credentials.get().
    authenticator_data (bytes): This attribute contains the authenticator data
      returned by the authenticator.
    signature (bytes): This attribute contains the raw signature returned from
      the authenticator.
    user_handle (Optional[bytes]): This attribute contains the user handle returned from
      the authenticator, or null if the authenticator did not return a user handle.

  References:
    * https://w3.org/TR/webauthn/#authenticatorassertionresponse
    * https://w3.org/TR/webauthn/#authenticatorresponse
  """
  def __init__(self,
               *,
               client_data_JSON: bytes,
               authenticator_data: bytes,
               signature: bytes,
               user_handle: Optional[bytes] = None):
    super().__init__(client_data_JSON=client_data_JSON)
    self.authenticator_data = authenticator_data
    self.signature = signature
    self.user_handle = user_handle


class Credential:
  """
  A credential is an object which allows a developer to make an authentication
  decision for a particular action.
  
  Attributes:
    id (str): The credential’s identifier.
      The requirements for the identifier are distinct for each type of
      credential. It might represent a username for username/password tuples,
      for example.
    type (str): Specifies the credential type represented by this object.

  References:
    * https://w3.org/TR/credential-management-1/#credential
  """
  def __init__(self, *, id: str, type: str):
    self.id = id
    self.type = type


class PublicKeyCredential(Credential):
  """
  PublicKeyCredential inherits from Credential, and contains the attributes
  that are returned to the caller when a new credential is created, or a new
  assertion is requested.

  Attributes:
    id (str): The credential’s identifier.
      The requirements for the identifier are distinct for each type of
      credential. It might represent a username for username/password tuples,
      for example.
    type (str): Specifies the credential type represented by this object.
    raw_id (bytes): This attribute is the raw credential id.
    response (AuthenticatorResponse): This attribute contains the
      authenticator's response to the client’s request to either create a
      public key credential, or generate an authentication assertion. If the
      PublicKeyCredential is created in response to
      navigator.credentials.create(), this attribute’s value will be an
      AuthenticatorAttestationResponse, otherwise, the PublicKeyCredential was
      created in response to navigator.credentials.get(), and this attribute’s
      value will be an AuthenticatorAssertionResponse.

  References:
    * https://w3.org/TR/webauthn/#iface-pkcredential
    * https://w3.org/TR/credential-management-1/#credential
  """
  def __init__(self, *, id: str, type: str, raw_id: bytes,
               response: AuthenticatorResponse):
    super().__init__(id=id, type=type)
    self.raw_id = raw_id
    self.response = response


class TokenBindingStatus(Enum):
  """
  The status of a Token Binding.

  Attributes:
    SUPPORTED (str):
      Indicates the client supports token binding, but it was not negotiated
      when communicating with the Relying Party.
    PRESENT (str):
      Indicates token binding was used when communicating with the Relying
      Party. In this case, the id member MUST be present.

  References:
    * https://w3.org/TR/webauthn/#enumdef-tokenbindingstatus
  """

  SUPPORTED = 'supported'
  PRESENT = 'present'


class TokenBinding:
  """
  TokenBinding contains information about the state of the Token Binding
  protocol used when the client was communicating with the Relying Party.

  Attributes:
    status (TokenBindingStatus): The status of the Token Binding.
    id (str): This member MUST be present if status is present, and MUST be a
      base64url encoding of the Token Binding ID that was used when
      the client was communicating with the Relying Party.

  References:
    * https://w3.org/TR/webauthn/#dictdef-tokenbinding
  """
  def __init__(self, *, status: TokenBindingStatus, id: Optional[str] = None):
    self.status = status
    self.id = id


class CollectedClientData:
  """
  The client data represents the contextual bindings of both the WebAuthn
  Relying Party and the client.

  Attributes:
    type (str): This member contains the string "webauthn.create" when creating
      new credentials, and "webauthn.get" when getting an assertion from an
      existing credential. The purpose of this member is to prevent certain
      types of signature confusion attacks (where an attacker substitutes one
      legitimate signature for another).
    challenge (str): This member contains the base64url encoding of the
      challenge provided by the Relying Party.
    origin (str): This member contains the fully qualified origin of the
      requester, as provided to the authenticator by the client, in the syntax
      defined by RFC6454.
    token_binding (Optional[TokenBinding]): This OPTIONAL member contains
      information about the state of the Token Binding protocol used when the
      client was communicating with the Relying Party. Its absence 
      that the client doesn’t support token binding.

  References:
    * https://w3.org/TR/webauthn/#dictdef-collectedclientdata
  """
  def __init__(self,
               *,
               type: str,
               challenge: str,
               origin: str,
               token_binding: Optional[TokenBinding] = None):
    self.type = type
    self.challenge = challenge
    self.origin = origin
    self.token_binding = token_binding


class AuthenticatorDataFlag(Enum):
  """
  The authenticator data bit flags.

  Attributes:
    UP (int): User is present.
    RFU1 (int): Reserved for future use.
    UV (int): User is verified.
    RFU2 (int): Reserved for future use.
    AT (int): Attested credential data included.
    ED (int): Extension data is included in the authenticator data.
    
  References:
    * https://w3.org/TR/webauthn/#sec-authenticator-data
  """

  UP = 1 << 0
  RFU1 = 1 << 1
  UV = 1 << 2
  RFU2 = (1 << 3) | (1 << 4) | (1 << 5)
  AT = 1 << 6
  ED = 1 << 7


class COSEKeyType(metaclass=NameValueEnumsContainer):
  """
  A metaclass for the COSEKeyOperation Name and Value Enums.
  
  Both the Name and Value Enums share the following documentation.

  The COSEKeyType object defines a way to hold a single key object. It is still
  required that the members of individual key types be defined.

  Attributes:
    OKP: Octet Key Pair
    EC2: Elliptic Curve Keys with X, Y Coordinate pair
    SYMMETRIC: Symmetric Keys
  
  References:
    * https://cose-wg.github.io/cose-spec/#rfc.section.13
  """
  class Name(Enum):
    OKP = 'OKP'
    EC2 = 'EC2'
    SYMMETRIC = 'Symmetric'

  class Value(Enum):
    OKP = 1
    EC2 = 2
    SYMMETRIC = 4


class COSEKeyOperation(metaclass=NameValueEnumsContainer):
  """
  A metaclass for the COSEKeyOperation Name and Value Enums.

  Both the Name and Value Enums share the following documentation.

  The COSEKeyOperation object specifies the type of operations that are
  permitted for the COSE Key.

  Attributes:
    SIGN: The key is used to create signatures. Requires private key fields.
    VERIFY: The key is used for verification of signatures.
    ENCRYPT: The key is used for key transport encryption.
    DECRYPT: The key is used for key transport decryption. Requires private key
      fields.
    WRAP_KEY: The key is used for key wrapping.
    UNWRAP_KEY: The key is used for key unwrapping. Requires private key
      fields.
    DERIVE_KEY: The key is used for deriving keys. Requires private key fields.
    DERIVE_BITS: The key is used for deriving bits not to be used as a key.
      Requires private key fields.
    MAC_CREATE: The key is used for creating MACs.
    MAC_VERIFY: The key is used for validating MACs.

  References:
    * https://cose-wg.github.io/cose-spec/#rfc.section.7.1
  """
  class Name(Enum):
    SIGN = 'sign'
    VERIFY = 'verify'
    ENCRYPT = 'encrypt'
    DECRYPT = 'decrypt'
    WRAP_KEY = 'wrap key'
    UNWRAP_KEY = 'unwrap key'
    DERIVE_KEY = 'derive key'
    DERIVE_BITS = 'derive bits'
    MAC_CREATE = 'MAC create'
    MAC_VERIFY = 'MAC verify'

  class Value(Enum):
    SIGN = 1
    VERIFY = 2
    ENCRYPT = 3
    DECRYPT = 4
    WRAP_KEY = 5
    UNWRAP_KEY = 6
    DERIVE_KEY = 7
    DERIVE_BITS = 8
    MAC_CREATE = 9
    MAC_VERIFY = 10


class EC2KeyType(metaclass=NameValueEnumsContainer):
  """
  A metaclass for the COSEKeyOperation Name and Value Enums.

  Both the Name and Value Enums share the following documentation.

  Attributes:
    P_256: NIST P-256 also known as secp256r1.
    P_384: NIST P-384 also known as secp384r1.
    P_521: NIST P-521 also known as secp521r1.

  References:
    * https://cose-wg.github.io/cose-spec/#rfc.section.13.1
  """
  class Name(Enum):
    P_256 = 'P-256'
    P_384 = 'P-384'
    P_521 = 'P-521'

  class Value(Enum):
    P_256 = 1
    P_384 = 2
    P_521 = 3


class OKPKeyType(metaclass=NameValueEnumsContainer):
  """
  A metaclass for the COSEKeyOperation Name and Value Enums.

  Both the Name and Value Enums share the following documentation.

  Attributes:
    ED25519: Ed25519 for use with EdDSA only.
    ED448: Ed448 for use with EdDSA only.

  References:
    * https://cose-wg.github.io/cose-spec/#rfc.section.13.1
  """
  class Name(Enum):
    ED25519 = 'Ed25519'
    ED448 = 'Ed448'

  class Value(Enum):
    ED25519 = 6
    ED448 = 7


class CredentialPublicKey:
  """
  The credential public key encoded in COSE_Key format.

  Attributes:
    kty (Union[COSEKeyType.Name, COSEKeyType.Value]):
      This parameter is used to identify the family of keys for this structure,
      and thus the set of key type specific parameters to be found. This 
      parameter MUST be present in a key object. Implementations MUST verify
      that the key type is appropriate for the algorithm being processed. The
      key type MUST be included as part of the trust decision process.
    kid (Optional[bytes]):
      This parameter is used to give an identifier for a key. The identifier is
      not structured and can be anything from a user provided string to a value
      computed on the public portion of the key. This field is intended for
      matching against a 'kid' parameter in a message in order to filter down
      the set of keys that need to be checked.
    alg (Optional[
      Union[COSEAlgorithmIdentifier.Name, COSEAlgorithmIdentifier.Value]]):
      This parameter is used to restrict the algorithm that is used with the
      key. If this parameter is present in the key structure, the application
      MUST verify that this algorithm matches the algorithm for which the key
      is being used. If the algorithms do not match, then this key object MUST
      NOT be used to perform the cryptographic operation. Note that the same
      key can be in a different key structure with a different or no algorithm
      specified, however this is considered to be a poor security practice.
    key_ops (Optional[
      Sequence[Union[COSEKeyOperation.Name, COSEKeyOperation.Value]]]):
      This parameter is defined to restrict the set of operations that a key is
      to be used for. Algorithms define the values of key ops that are
      permitted to appear and are required for specific operations.
    base_IV (Optional[bytes]): 
      This parameter is defined to carry the base portion of an IV. This field
      provides the ability to associate a partial IV with a key that is then
      modified on a per message basis with the partial IV. Extreme care needs
      to be taken when using a Base IV in an application. Many encryption
      algorithms lose security if the same IV is used twice.

  References:
    * https://w3.org/TR/webauthn/#sec-attested-credential-data
    * https://cose-wg.github.io/cose-spec/#rfc.section.7
    * https://cose-wg.github.io/cose-spec/#rfc.section.7.1
  """
  def __init__(
      self,
      *,
      kty: Union[COSEKeyType.Name, COSEKeyType.Value],
      kid: Optional[bytes] = None,
      alg: Optional[Union[COSEAlgorithmIdentifier.Name,
                          COSEAlgorithmIdentifier.Value]] = None,
      key_ops: Optional[Sequence[Union[COSEKeyOperation.Name,
                                       COSEKeyOperation.Value]]] = None,
      base_IV: Optional[bytes] = None):
    self.kty = kty
    self.kid = kid
    self.alg = alg
    self.key_ops = key_ops
    self.base_IV = base_IV


class EC2CredentialPublicKey(CredentialPublicKey):
  """
  The Elliptic Curve Cryptography (EC2) credential public key encoded in
  COSE_Key format.

  Attributes:
    kty (Union[COSEKeyType.Name, COSEKeyType.Value]):
      This parameter is used to identify the family of keys for this structure,
      and thus the set of key type specific parameters to be found. This 
      parameter MUST be present in a key object. Implementations MUST verify
      that the key type is appropriate for the algorithm being processed. The
      key type MUST be included as part of the trust decision process.
    kid (Optional[bytes]):
      This parameter is used to give an identifier for a key. The identifier is
      not structured and can be anything from a user provided string to a value
      computed on the public portion of the key. This field is intended for
      matching against a 'kid' parameter in a message in order to filter down
      the set of keys that need to be checked.
    alg (Optional[
      Union[COSEAlgorithmIdentifier.Name, COSEAlgorithmIdentifier.Value]]):
      This parameter is used to restrict the algorithm that is used with the
      key. If this parameter is present in the key structure, the application
      MUST verify that this algorithm matches the algorithm for which the key
      is being used. If the algorithms do not match, then this key object MUST
      NOT be used to perform the cryptographic operation. Note that the same
      key can be in a different key structure with a different or no algorithm
      specified, however this is considered to be a poor security practice.
    key_ops (Optional[
      Sequence[Union[COSEKeyOperation.Name, COSEKeyOperation.Value]]]):
      This parameter is defined to restrict the set of operations that a key is
      to be used for. Algorithms define the values of key ops that are
      permitted to appear and are required for specific operations.
    base_IV (Optional[bytes]): 
      his parameter is defined to carry the base portion of an IV. This field
      provides the ability to associate a partial IV with a key that is then
      modified on a per message basis with the partial IV. Extreme care needs
      to be taken when using a Base IV in an application. Many encryption
      algorithms lose security if the same IV is used twice.
    x (bytes): The Elliptic Curve X Coordinate.
    y (bytes): The Elliptic Curve Y Coordinate.
    crv (Union[EC2KeyType.Name, EC2KeyType.Value]):
      The Elliptic Curve Identifier.

  References:
    * https://w3.org/TR/webauthn/#sec-attested-credential-data
    * https://cose-wg.github.io/cose-spec/#rfc.section.7
    * https://cose-wg.github.io/cose-spec/#rfc.section.7.1
    * https://cose-wg.github.io/cose-spec/#rfc.section.13.1
    * https://cose-wg.github.io/cose-spec/#rfc.section.13.1.1
  """
  def __init__(
      self,
      *,
      kty: Union[COSEKeyType.Name, COSEKeyType.Value],
      kid: Optional[bytes] = None,
      alg: Optional[Union[COSEAlgorithmIdentifier.Name,
                          COSEAlgorithmIdentifier.Value]] = None,
      key_ops: Optional[Sequence[Union[COSEKeyOperation.Name,
                                       COSEKeyOperation.Value]]] = None,
      base_IV: Optional[bytes] = None,
      x: bytes,
      y: bytes,
      crv: Union[EC2KeyType.Name, EC2KeyType.Value]):
    super().__init__(kty=kty,
                     kid=kid,
                     alg=alg,
                     key_ops=key_ops,
                     base_IV=base_IV)
    self.x = x
    self.y = y
    self.crv = crv


class OKPCredentialPublicKey(CredentialPublicKey):
  """
  The Octet Key Pair (OKP) credential public key encoded in COSE_Key format.

  Attributes:
    kty (Union[COSEKeyType.Name, COSEKeyType.Value]):
      This parameter is used to identify the family of keys for this structure,
      and thus the set of key type specific parameters to be found. This 
      parameter MUST be present in a key object. Implementations MUST verify
      that the key type is appropriate for the algorithm being processed. The
      key type MUST be included as part of the trust decision process.
    kid (Optional[bytes]):
      This parameter is used to give an identifier for a key. The identifier is
      not structured and can be anything from a user provided string to a value
      computed on the public portion of the key. This field is intended for
      matching against a 'kid' parameter in a message in order to filter down
      the set of keys that need to be checked.
    alg (Optional[
      Union[COSEAlgorithmIdentifier.Name, COSEAlgorithmIdentifier.Value]]):
      This parameter is used to restrict the algorithm that is used with the
      key. If this parameter is present in the key structure, the application
      MUST verify that this algorithm matches the algorithm for which the key
      is being used. If the algorithms do not match, then this key object MUST
      NOT be used to perform the cryptographic operation. Note that the same
      key can be in a different key structure with a different or no algorithm
      specified, however this is considered to be a poor security practice.
    key_ops (Optional[
      Sequence[Union[COSEKeyOperation.Name, COSEKeyOperation.Value]]]):
      This parameter is defined to restrict the set of operations that a key is
      to be used for. Algorithms define the values of key ops that are
      permitted to appear and are required for specific operations.
    base_IV (Optional[bytes]): 
      his parameter is defined to carry the base portion of an IV. This field
      provides the ability to associate a partial IV with a key that is then
      modified on a per message basis with the partial IV. Extreme care needs
      to be taken when using a Base IV in an application. Many encryption
      algorithms lose security if the same IV is used twice.
    x (bytes): The Elliptic Curve X Coordinate.
    y (bytes): The Elliptic Curve Y Coordinate.
    crv (Union[OKPKeyType.Name, OKPKeyType.Value]):
      The Elliptic Curve Identifier.

  References:
    * https://w3.org/TR/webauthn/#sec-attested-credential-data
    * https://cose-wg.github.io/cose-spec/#rfc.section.7
    * https://cose-wg.github.io/cose-spec/#rfc.section.7.1
    * https://cose-wg.github.io/cose-spec/#rfc.section.13.1
    * https://cose-wg.github.io/cose-spec/#rfc.section.13.2
  """
  def __init__(
      self,
      *,
      kty: Union[COSEKeyType.Name, COSEKeyType.Value],
      kid: Optional[bytes] = None,
      alg: Optional[Union[COSEAlgorithmIdentifier.Name,
                          COSEAlgorithmIdentifier.Value]] = None,
      key_ops: Optional[Sequence[Union[COSEKeyOperation.Name,
                                       COSEKeyOperation.Value]]] = None,
      base_IV: Optional[bytes] = None,
      crv: Union[OKPKeyType.Name, OKPKeyType.Value],
      x: bytes):
    super().__init__(kty=kty,
                     kid=kid,
                     alg=alg,
                     key_ops=key_ops,
                     base_IV=base_IV)
    self.crv = crv
    self.x = x


class AttestedCredentialData:
  """
  Attested credential data is a variable-length byte array added to the
  authenticator data when generating an attestation object for a given
  credential.

  Attributes:
    aaguid (bytes): The AAGUID of the authenticator.
      Each authenticator has an AAGUID, which is a 128-bit identifier
      indicating the type (e.g. make and model) of the authenticator. The
      AAGUID MUST be chosen by the manufacturer to be identical across all
      substantially identical authenticators made by that manufacturer, and
      different (with high probability) from the AAGUIDs of all other types of
      authenticators. The AAGUID for a given type of authenticator SHOULD be
      randomly generated to ensure this. The Relying Party MAY use the AAGUID
      to infer certain properties of the authenticator, such as certification
      level and strength of key protection, using information from other
      sources.
    credential_id_length (int): Byte length of Credential ID, parsed from a
      16-bit unsigned big-endian integer.
    credential_id (bytes):
      A probabilistically-unique byte sequence identifying a public key
      credential source and its authentication assertions.
    credential_public_key (Optional[CredentialPublicKey]):
      The credential public key encoded in COSE_Key format. The
      COSE_Key-encoded credential public key MUST contain the "alg" parameter
      and MUST NOT contain any other OPTIONAL parameters. The "alg" parameter
      MUST contain a COSEAlgorithmIdentifier value. The encoded credential
      public key MUST also contain any additional REQUIRED parameters
      stipulated by the relevant key type specification, i.e., REQUIRED for the
      key type "kty" and algorithm "alg"

  References:
    * https://w3.org/TR/webauthn/#sec-attested-credential-data
    * https://w3.org/TR/webauthn/#aaguid
    * https://w3.org/TR/webauthn/#credential-id
    * 
  """
  def __init__(self,
               aaguid: bytes,
               credential_id_length: int,
               credential_id: bytes,
               credential_public_key: Optional[CredentialPublicKey] = None):
    self.aaguid = aaguid
    self.credential_id_length = credential_id_length
    self.credential_id = credential_id
    self.credential_public_key = credential_public_key


class AuthenticatorData:
  """
  The authenticator data structure encodes contextual bindings made by the
  authenticator. These bindings are controlled by the authenticator itself, and
  derive their trust from the WebAuthn Relying Party's assessment of the
  security properties of the authenticator. In one extreme case, the
  authenticator may be embedded in the client, and its bindings may be no more
  trustworthy than the client data. At the other extreme, the authenticator may
  be a discrete entity with high-security hardware and software, connected to
  the client over a secure channel. In both cases, the Relying Party receives
  the authenticator data in the same format, and uses its knowledge of the
  authenticator to make trust decisions.

  Attributes:
    rp_id_hash (bytes): SHA-256 hash of the RP ID the credential is scoped to.
    flags (int): 8 flag bits for the authenticator data.
    sign_count (int):
      Signature counter parsed from a 32-bit unsigned big-endian integer.
    attested_credential_data (Optional[AttestedCredentialData]):
      The attested credential data.
    extensions (Optional[AuthenticationExtensionsClientOutputs]):
      Extension-defined authenticator data.

  References:
    * https://w3.org/TR/webauthn/#authenticator-data
  """
  def __init__(
      self,
      *,
      rp_id_hash: bytes,
      flags: int,
      sign_count: int,
      attested_credential_data: Optional[AttestedCredentialData] = None,
      extensions: Optional[AuthenticationExtensionsClientOutputs] = None):
    self.rp_id_hash = rp_id_hash
    self.flags = flags
    self.sign_count = sign_count
    self.attested_credential_data = attested_credential_data
    self.extensions = extensions


class AttestationStatementFormatIdentifier(Enum):
  """
  Attestation statement formats are identified by a string, called an
  attestation statement format identifier, chosen by the author of the
  attestation statement format.

  Attributes:
    PACKED (str):
      This is a WebAuthn optimized attestation statement format. It uses a very
      compact but still extensible encoding method. It is implementable by
      authenticators with limited resources (e.g., secure elements).
    TPM (str):
      This attestation statement format is generally used by authenticators
      that use a Trusted Platform Module as their cryptographic engine.
    ANDROID_KEY (str):
      When the authenticator in question is a platform-provided Authenticator
      on the Android "N" or later platform, the attestation statement is based
      on the Android key attestation. In these cases, the attestation statement
      is produced by a component running in a secure operating environment, but
      the authenticator data for the attestation is produced outside this
      environment. The WebAuthn Relying Party is expected to check that the
      authenticator data claimed to have been used for the attestation is
      consistent with the fields of the attestation certificate’s extension
      data.
    ANDROID_SAFETYNET (str):
      When the authenticator in question is a platform-provided Authenticator
      on certain Android platforms, the attestation statement is based on the
      SafetyNet API. In this case the authenticator data is completely
      controlled by the caller of the SafetyNet API (typically an application
      running on the Android platform) and the attestation statement only
      provides some statements about the health of the platform and the
      identity of the calling application. This attestation does not provide
      information regarding provenance of the authenticator and its associated
      data. Therefore platform-provided authenticators SHOULD make use of the
      Android Key Attestation when available, even if the SafetyNet API is also
      present.
    FIDO_U2F (str):
      This attestation statement format is used with FIDO U2F authenticators.
    NONE (str):
      The none attestation statement format is used to replace any
      authenticator-provided attestation statement when a WebAuthn Relying
      Party indicates it does not wish to receive attestation information.

  References:
    * https://w3.org/TR/webauthn/#attestation-statement-format-identifier
  """

  PACKED = 'packed'
  TPM = 'tpm'
  ANDROID_KEY = 'android-key'
  ANDROID_SAFETYNET = 'android-safetynet'
  FIDO_U2F = 'fido-u2f'
  NONE = 'none'


class AttestationType(Enum):
  """
  WebAuthn supports several attestation types, defining the semantics of
  attestation statements and their underlying trust models.

  Attributes:
    BASIC (str):
      In the case of basic attestation, the authenticator’s attestation key pair
      is specific to an authenticator model. Thus, authenticators of the same
      model often share the same attestation key pair.
    SELF (str):
      In the case of self attestation, also known as surrogate basic attestation,
      the Authenticator does not have any specific attestation key. Instead it
      uses the credential private key to create the attestation signature.
      Authenticators without meaningful protection measures for an attestation
      private key typically use this attestation type.
    ATTCA (str):
      In this case, an authenticator is based on a Trusted Platform Module (TPM)
      and holds an authenticator-specific "endorsement key" (EK). This key is
      used to securely communicate with a trusted third party, the Attestation CA
      (formerly known as a "Privacy CA"). The authenticator can generate multiple
      attestation identity key pairs (AIK) and requests an Attestation CA to
      issue an AIK certificate for each. Using this approach, such an
      authenticator can limit the exposure of the EK (which is a global
      correlation handle) to Attestation CA(s). AIKs can be requested for each
      authenticator-generated public key credential individually, and conveyed
      to Relying Parties as attestation certificates.
    ECDAA (str):
      In this case, the Authenticator receives direct anonymous attestation (DAA)
      credentials from a single DAA-Issuer. These DAA credentials are used along
      with blinding to sign the attested credential data. The concept of blinding
      avoids the DAA credentials being misused as global correlation handle.
      WebAuthn supports DAA using elliptic curve cryptography and bilinear
      pairings, called ECDAA in this specification. Consequently we denote the
      DAA-Issuer as ECDAA-Issuer.
    NONE (str): In this case, no attestation information is available.
    UNCERTAIN (str): In this case, the attestation type is uncertain.

  References:
    * https://w3.org/TR/webauthn/#sctn-attestation-types
  """

  BASIC = 'Basic'
  ATTCA = 'AttCA'
  ECDAA = 'ECDAA'
  SELF = 'Self'
  NONE = 'None'
  UNCERTAIN = 'Uncertain'


class AttestationStatement:
  """
  An important component of the attestation object is the attestation
  statement. This is a specific type of signed data object, containing
  statements about a public key credential itself and the authenticator that
  created it. It contains an attestation signature created using the key of the
  attesting authority (except for the case of self attestation, when it is
  created using the credential private key). In order to correctly interpret an
  attestation statement, a Relying Party needs to understand these two aspects
  of attestation:

    1. The attestation statement format is the manner in which the signature is
    represented and the various contextual bindings are incorporated into the
    attestation statement by the authenticator. In other words, this defines
    the syntax of the statement. Various existing components and OS platforms
    (such as TPMs and the Android OS) have previously defined attestation
    statement formats.
    2. The attestation type defines the semantics of attestation statements and
    their underlying trust models. Specifically, it defines how a Relying Party
    establishes trust in a particular attestation statement, after verifying
    that it is cryptographically valid.

  Attributes:
    alg (Optional[COSEAlgorithmIdentifier]):
      A COSEAlgorithmIdentifier containing the identifier of the algorithm used
      to generate the attestation signature.
    sig (Optional[bytes]): A byte string containing the attestation signature.

  References:
    * https://w3.org/TR/webauthn/#attestation-statement
  """
  def __init__(self,
               *,
               alg: Optional[Union[COSEAlgorithmIdentifier.Name,
                                   COSEAlgorithmIdentifier.Value]] = None,
               sig: Optional[bytes] = None):
    self.alg = alg
    self.sig = sig


class PackedAttestationStatement(AttestationStatement):
  """
  The Packed Attestation Statement.

  Attributes:
    alg (COSEAlgorithmIdentifier):
      A COSEAlgorithmIdentifier containing the identifier of the algorithm used
      to generate the attestation signature.
    sig (bytes): A byte string containing the attestation signature.

  References:
    * https://w3.org/TR/webauthn/#attestation-statement
    * https://w3.org/TR/webauthn/#packed-attestation
  """
  def __init__(self, *, alg: Union[COSEAlgorithmIdentifier.Name,
                                   COSEAlgorithmIdentifier.Value], sig: bytes):
    super().__init__(alg=alg, sig=sig)


class PackedX509AttestationStatement(PackedAttestationStatement):
  """
  The Packed X.509 Attestation Statement.

  Attributes:
    alg (COSEAlgorithmIdentifier):
      A COSEAlgorithmIdentifier containing the identifier of the algorithm used
      to generate the attestation signature.
    sig (bytes): A byte string containing the attestation signature.
    x5c (Sequence[bytes]):
      The elements of this array contain the attestation certificate and its 
      certificate chain, each encoded in X.509 format. The attestation
      certificate MUST be the first element in the array.

  References:
    * https://w3.org/TR/webauthn/#attestation-statement
    * https://w3.org/TR/webauthn/#packed-attestation
  """
  def __init__(self, *, alg: Union[COSEAlgorithmIdentifier.Name,
                                   COSEAlgorithmIdentifier.Value], sig: bytes,
               x5c: Sequence[bytes]):
    super().__init__(alg=alg, sig=sig)
    self.x5c = x5c


class PackedECDAAAttestationStatement(PackedAttestationStatement):
  """
  The Packed ECDAA Attestation Statement.

  Attributes:
    alg (COSEAlgorithmIdentifier):
      A COSEAlgorithmIdentifier containing the identifier of the algorithm used
      to generate the attestation signature.
    sig (bytes): A byte string containing the attestation signature.
    ecdaa_key_id (bytes):
      The identifier of the ECDAA-Issuer public key. This is the BigNumberToB
      encoding of the component "c" of the ECDAA-Issuer public key.

  References:
    * https://w3.org/TR/webauthn/#attestation-statement
    * https://w3.org/TR/webauthn/#packed-attestation
  """
  def __init__(self, *, alg: Union[COSEAlgorithmIdentifier.Name,
                                   COSEAlgorithmIdentifier.Value], sig: bytes,
               ecdaa_key_id: bytes):
    super().__init__(alg=alg, sig=sig)
    self.ecdaa_key_id = ecdaa_key_id


class TPMAttestationStatement(AttestationStatement):
  """
  The TPM Attestation Statement.

  Attributes:
    alg (COSEAlgorithmIdentifier):
      A COSEAlgorithmIdentifier containing the identifier of the algorithm used
      to generate the attestation signature.
    sig (bytes): A byte string containing the attestation signature.
    ver (str):
      The version of the TPM specification to which the signature conforms.
    cert_info (bytes):
      The TPMS_ATTEST structure over which the above signature was computed.
    pub_area (bytes):
      The TPMT_PUBLIC structure used by the TPM to represent the credential
      public key.

  References:
    * https://w3.org/TR/webauthn/#attestation-statement
    * https://w3.org/TR/webauthn/#tpm-attestation
  """
  def __init__(self, *, alg: Union[COSEAlgorithmIdentifier.Name,
                                   COSEAlgorithmIdentifier.Value], sig: bytes,
               ver: str, cert_info: bytes, pub_area: bytes):
    super().__init__(alg=alg, sig=sig)
    self.ver = ver
    self.cert_info = cert_info
    self.pub_area = pub_area


class TPMX509AttestationStatement(TPMAttestationStatement):
  """
  The TPM X.509 Attestation Statement.

  Attributes:
    alg (COSEAlgorithmIdentifier):
      A COSEAlgorithmIdentifier containing the identifier of the algorithm used
      to generate the attestation signature.
    sig (bytes): A byte string containing the attestation signature.
    ver (str):
      The version of the TPM specification to which the signature conforms.
    cert_info (bytes):
      The TPMS_ATTEST structure over which the above signature was computed.
    pub_area (bytes):
      The TPMT_PUBLIC structure used by the TPM to represent the credential
      public key.
    x5c (Sequence[bytes]):
      The elements of this array contain the attestation certificate and its 
      certificate chain, each encoded in X.509 format. The attestation
      certificate MUST be the first element in the array.

  References:
    * https://w3.org/TR/webauthn/#attestation-statement
    * https://w3.org/TR/webauthn/#tpm-attestation
  """
  def __init__(self, *, alg: Union[COSEAlgorithmIdentifier.Name,
                                   COSEAlgorithmIdentifier.Value], sig: bytes,
               ver: str, cert_info: bytes, pub_area: bytes,
               x5c: Sequence[bytes]):
    super().__init__(alg=alg,
                     sig=sig,
                     ver=ver,
                     cert_info=cert_info,
                     pub_area=pub_area)
    self.x5c = x5c


class TPMECDAAAttestationStatement(TPMAttestationStatement):
  """
  The TPM ECDAA Attestation Statement.

  Attributes:
    alg (COSEAlgorithmIdentifier):
      A COSEAlgorithmIdentifier containing the identifier of the algorithm used
      to generate the attestation signature.
    sig (bytes): A byte string containing the attestation signature.
    ver (str):
      The version of the TPM specification to which the signature conforms.
    cert_info (bytes):
      The TPMS_ATTEST structure over which the above signature was computed.
    pub_area (bytes):
      The TPMT_PUBLIC structure used by the TPM to represent the credential
      public key.
    ecdaa_key_id (bytes):
      The identifier of the ECDAA-Issuer public key. This is the BigNumberToB
      encoding of the component "c" of the ECDAA-Issuer public key.

  References:
    * https://w3.org/TR/webauthn/#attestation-statement
    * https://w3.org/TR/webauthn/#tpm-attestation
  """
  def __init__(self, *, alg: Union[COSEAlgorithmIdentifier.Name,
                                   COSEAlgorithmIdentifier.Value], sig: bytes,
               ver: str, cert_info: bytes, pub_area: bytes,
               ecdaa_key_id: bytes):
    super().__init__(alg=alg,
                     sig=sig,
                     ver=ver,
                     cert_info=cert_info,
                     pub_area=pub_area)
    self.ecdaa_key_id = ecdaa_key_id


class AndroidKeyAttestationStatement(AttestationStatement):
  """
  The Android Key Attestation Statement.

  Attributes:
    alg (COSEAlgorithmIdentifier):
      A COSEAlgorithmIdentifier containing the identifier of the algorithm used
      to generate the attestation signature.
    sig (bytes): A byte string containing the attestation signature.
    x5c (Sequence[bytes]):
      The elements of this array contain the attestation certificate and its 
      certificate chain, each encoded in X.509 format. The attestation
      certificate MUST be the first element in the array.

  References:
    * https://w3.org/TR/webauthn/#attestation-statement
    * https://w3.org/TR/webauthn/#android-key-attestation
  """
  def __init__(self, *, alg: Union[COSEAlgorithmIdentifier.Name,
                                   COSEAlgorithmIdentifier.Value], sig: bytes,
               x5c: Sequence[bytes]):
    super().__init__(alg=alg, sig=sig)
    self.x5c = x5c


class AndroidSafetyNetAttestationStatement(AttestationStatement):
  """
  The Android SafetyNet Attestation Statement.

  Attributes:
    ver (str):
      The version number of Google Play Services responsible for providing the
      SafetyNet API.
    response (bytes):
      The UTF-8 encoded result of the getJwsResult() call of the SafetyNet API.
      This value is a JWS object (see SafetyNet online documentation) in
      Compact Serialization.

  References:
    * https://w3.org/TR/webauthn/#attestation-statement
    * https://w3.org/TR/webauthn/#android-safetynet-attestation
  """
  def __init__(self, *, ver: str, response: bytes):
    super().__init__(alg=None, sig=None)
    self.ver = ver
    self.response = response


class FIDOU2FAttestationStatement(AttestationStatement):
  """
  The FIDO U2F Attestation Statement.

  Attributes:
    sig (bytes):
      The attestation signature. The signature was calculated over the (raw)
      U2F registration response message received by the client from the
      authenticator.
    x5c (Sequence[bytes]):
      A single element array containing the attestation certificate in X.509
      format.

  References:
    * https://w3.org/TR/webauthn/#attestation-statement
    * https://w3.org/TR/webauthn/#android-safetynet-attestation
  """
  def __init__(self, *, sig: bytes, x5c: Sequence[bytes]):
    super().__init__(alg=None, sig=sig)
    self.x5c = x5c


class NoneAttestationStatement(AttestationStatement):
  """
  The None Attestation Statement. Consists of an empty object.

  References:
    * https://w3.org/TR/webauthn/#none-attestation
  """
  pass


class AttestationObject:
  """
  Authenticators MUST also provide some form of attestation. The basic
  requirement is that the authenticator can produce, for each credential public
  key, an attestation statement verifiable by the WebAuthn Relying Party.
  Typically, this attestation statement contains a signature by an attestation
  private key over the attested credential public key and a challenge, as well
  as a certificate or similar data providing provenance information for the
  attestation public key, enabling the Relying Party to make a trust decision.
  However, if an attestation key pair is not available, then the authenticator
  MUST perform self attestation of the credential public key with the
  corresponding credential private key. All this information is returned by
  authenticators any time a new public key credential is generated, in the
  overall form of an attestation object.

  Attributes:
    auth_data (AuthenticatorData): The authenticator data.
    fmt (AttestationStatementFormatIdentifier):
      The format of the attestation statement.
    att_stmt (AttestationStatment): The attestation statement.

  References:
    * https://w3.org/TR/webauthn/#attestation-object
  """
  def __init__(self, *, auth_data: AuthenticatorData,
               fmt: AttestationStatementFormatIdentifier,
               att_stmt: AttestationStatement):
    self.auth_data = auth_data
    self.fmt = fmt
    self.att_stmt = att_stmt



import base64
import binascii
import io
import json
import struct
from enum import Enum
from functools import singledispatch
from typing import Any, List, Optional, Sequence, Set, Tuple, Union, cast

import cbor

from webauthn_rp.constants import (EC2_P_256_NUMBER_LENGTH,
                                   EC2_P_384_NUMBER_LENGTH,
                                   EC2_P_521_NUMBER_LENGTH,
                                   OKP_ED448_NUMBER_LENGTH,
                                   OKP_ED25519_NUMBER_LENGTH)
from webauthn_rp.errors import (DecodingError, TokenBindingError,
                                ValidationError)
from webauthn_rp.types import (
    AndroidKeyAttestationStatement, AndroidSafetyNetAttestationStatement,
    AttestationObject, AttestationStatement,
    AttestationStatementFormatIdentifier, AttestedCredentialData,
    AuthenticationExtensionsClientOutputs, AuthenticatorAssertionResponse,
    AuthenticatorAttestationResponse, AuthenticatorData, CollectedClientData,
    Coordinates, COSEAlgorithmIdentifier, COSEKeyOperation, COSEKeyType,
    CredentialPublicKey, EC2CredentialPublicKey, EC2KeyType,
    FIDOU2FAttestationStatement, NoneAttestationStatement,
    OKPCredentialPublicKey, OKPKeyType, PackedAttestationStatement,
    PackedECDAAAttestationStatement, PackedX509AttestationStatement,
    PublicKeyCredential, TokenBinding, TokenBindingStatus,
    TPMAttestationStatement, TPMECDAAAttestationStatement,
    TPMX509AttestationStatement)
from webauthn_rp.validators import validate


def parse_dictionary_field(field_key: Any,
                           valid_types: Union[type, Sequence[type]],
                           dictionary: dict,
                           required: bool = True) -> Any:
  valid_types_seq: Sequence[type] = [valid_types] if (  # type: ignore
      type(valid_types) is type) else valid_types
  field = dictionary.get(field_key)
  if field is None:
    if not required: return
    raise ValidationError(
        '{} is required in dictionary keys'.format(field_key))

  if type(field) not in valid_types_seq:
    raise ValidationError('{} type must be one of {} not {}'.format(
        field_key, str(valid_types_seq), str(type(field))))

  return field


def check_unsupported_keys(supported: Set[str], data: dict):
  unsupported_keys = set(data.keys()).difference(supported)
  if unsupported_keys:
    raise ValidationError(
        ('Found unsupported keys in data {}').format(str(unsupported_keys)))


def bytes_from_base64(s: str) -> bytes:
  try:
    return base64.b64decode(s)
  except Exception:
    raise ValidationError('Invalid base64 string')


def bytes_from_array(arr: Sequence[Any]) -> bytes:
  try:
    return bytes(arr)
  except binascii.Error:
    raise ValidationError('Invalid array type')


def parse_public_key_credential(data: dict) -> PublicKeyCredential:
  check_unsupported_keys({'id', 'rawId', 'response', 'type'}, data)
  id_ = parse_dictionary_field('id', str, data)
  type_ = parse_dictionary_field('type', str, data)
  raw_id = bytes_from_base64(parse_dictionary_field('rawId', str, data))

  response = parse_dictionary_field('response', dict, data)
  client_data_JSON = bytes_from_base64(
      parse_dictionary_field('clientDataJSON', str, response))

  if 'attestationObject' in response:
    # Parse AuthenticatorAttestationResponse
    check_unsupported_keys({'clientDataJSON', 'attestationObject'}, response)

    attestation_object = bytes_from_base64(
        parse_dictionary_field('attestationObject', str, response))

    return PublicKeyCredential(id=id_,
                               type=type_,
                               raw_id=raw_id,
                               response=AuthenticatorAttestationResponse(
                                   client_data_JSON=client_data_JSON,
                                   attestation_object=attestation_object))
  else:
    # Parse AuthenticatorAssertionResponse
    check_unsupported_keys(
        {'clientDataJSON', 'authenticatorData', 'signature', 'userHandle'},
        response)

    authenticator_data = bytes_from_base64(
        parse_dictionary_field('authenticatorData', str, response))
    signature = bytes_from_base64(
        parse_dictionary_field('signature', str, response))
    user_handle_b64s = parse_dictionary_field('userHandle', (list, tuple),
                                              response, False)
    user_handle = None
    if user_handle_b64s is not None:
      user_handle = bytes_from_base64(user_handle_b64s)

    return PublicKeyCredential(id=id_,
                               type=type_,
                               raw_id=raw_id,
                               response=AuthenticatorAssertionResponse(
                                   client_data_JSON=client_data_JSON,
                                   authenticator_data=authenticator_data,
                                   signature=signature,
                                   user_handle=user_handle))


def parse_credential_public_key_kty(
    credential_public_key: dict) -> Union[COSEKeyType.Name, COSEKeyType.Value]:
  kty_raw = parse_dictionary_field(1, (int, str), credential_public_key)

  try:
    kty = cast(Union[COSEKeyType.Name, COSEKeyType.Value],
               COSEKeyType(kty_raw))  # type: ignore
  except KeyError:
    raise ValidationError(
        'Invalid credential public key type {}'.format(kty_raw))
  return kty


def parse_credential_public_key_alg(
    credential_public_key: dict
) -> Union[COSEAlgorithmIdentifier.Name, COSEAlgorithmIdentifier.Value]:
  alg_raw = parse_dictionary_field(3, (int, str), credential_public_key)

  try:
    alg = cast(Union[COSEAlgorithmIdentifier.Name,
                     COSEAlgorithmIdentifier.Value],
               COSEAlgorithmIdentifier(alg_raw))  # type: ignore
  except KeyError:
    raise ValidationError(
        'Invalid credential public key alg type {}'.format(alg_raw))
  return alg


def parse_credential_public_key_key_ops(
    credential_public_key: dict
) -> Optional[Sequence[Union[COSEKeyOperation.Name, COSEKeyOperation.Value]]]:
  key_ops_raw = parse_dictionary_field(4, (list, tuple), credential_public_key,
                                       False)

  if key_ops_raw is None: return None

  if len(key_ops_raw) < 1:
    raise ValidationError(
        'Credential public key key_ops(4) must have at least 1 element')

  key_ops: List[Union[COSEKeyOperation.Name, COSEKeyOperation.Value]] = []
  for i, ko in enumerate(key_ops_raw):
    if type(ko) not in (str, int):
      raise ValidationError(
          ('Credential public key key_ops(3) index {} should either be a'
           ' text string or an integer').format(i))

    try:
      key_ops.append(COSEKeyOperation(ko))  # type: ignore
    except KeyError:
      raise ValidationError('Invalid credential public key key op {}'.format(
          credential_public_key[4]))

  return key_ops


def parse_credential_public_key_kwargs(credential_public_key: dict) -> dict:
  kty = parse_credential_public_key_kty(credential_public_key)
  kid = parse_dictionary_field(2, bytes, credential_public_key, False)
  alg = parse_credential_public_key_alg(credential_public_key)
  key_ops = parse_credential_public_key_key_ops(credential_public_key)
  base_IV = parse_dictionary_field(5, bytes, credential_public_key, False)
  return {
      'kty': kty,
      'kid': kid,
      'alg': alg,
      'key_ops': key_ops,
      'base_IV': base_IV
  }


def ec2_number_length(crv: Union[EC2KeyType.Name, EC2KeyType.Value]) -> int:
  key_to_length = {
      EC2KeyType.Name.P_256: EC2_P_256_NUMBER_LENGTH,
      EC2KeyType.Name.P_384: EC2_P_384_NUMBER_LENGTH,
      EC2KeyType.Name.P_521: EC2_P_521_NUMBER_LENGTH,
      EC2KeyType.Value.P_256: EC2_P_256_NUMBER_LENGTH,
      EC2KeyType.Value.P_384: EC2_P_384_NUMBER_LENGTH,
      EC2KeyType.Value.P_521: EC2_P_521_NUMBER_LENGTH,
  }

  return key_to_length[crv]


def okp_number_length(crv: Union[OKPKeyType.Name, OKPKeyType.Value]) -> int:
  key_to_length = {
      OKPKeyType.Name.ED25519: OKP_ED25519_NUMBER_LENGTH,
      OKPKeyType.Name.ED448: OKP_ED448_NUMBER_LENGTH,
      OKPKeyType.Value.ED25519: OKP_ED25519_NUMBER_LENGTH,
      OKPKeyType.Value.ED448: OKP_ED448_NUMBER_LENGTH,
  }

  return key_to_length[crv]


def parse_ec2_public_key_crv(
    credential_public_key: dict) -> Union[EC2KeyType.Name, EC2KeyType.Value]:
  crv_raw = parse_dictionary_field(-1, (int, str), credential_public_key)
  try:
    return EC2KeyType(crv_raw)  # type: ignore
  except KeyError:
    raise ValidationError('Invalid EC2 curve {}'.format(crv_raw))


def parse_okp_public_key_crv(
    credential_public_key: dict) -> Union[OKPKeyType.Name, OKPKeyType.Value]:
  crv_raw = parse_dictionary_field(-1, (int, str), credential_public_key)
  try:
    return OKPKeyType(crv_raw)  # type: ignore
  except KeyError:
    raise ValidationError('Invalid OKP curve {}'.format(crv_raw))


def parse_okp_public_key(credential_public_key: dict) -> CredentialPublicKey:
  x = parse_dictionary_field(-2, bytes, credential_public_key)
  crv = parse_okp_public_key_crv(credential_public_key)
  crv_len = okp_number_length(crv)
  if len(x) != crv_len:
    raise ValidationError(
        'Packed credential public key x and y must be {} bytes'.format(
            crv_len))

  return OKPCredentialPublicKey(
      crv=crv,
      x=x,
      **parse_credential_public_key_kwargs(credential_public_key),
  )


def parse_ec2_public_key(
    credential_public_key: dict) -> EC2CredentialPublicKey:
  x = parse_dictionary_field(-2, bytes, credential_public_key)
  y = parse_dictionary_field(-3, bytes, credential_public_key)
  crv = parse_ec2_public_key_crv(credential_public_key)
  crv_len = ec2_number_length(crv)
  if len(x) != crv_len or len(y) != crv_len:
    raise ValidationError(
        'Packed credential public key x and y must be {} bytes'.format(
            crv_len))

  return EC2CredentialPublicKey(
      x=x,
      y=y,
      crv=crv,
      **parse_credential_public_key_kwargs(credential_public_key),
  )


class CredentialPublicKeyParser(Enum):
  OKP = parse_okp_public_key
  EC2 = parse_ec2_public_key


def parse_extensions(
    extensions: dict) -> AuthenticationExtensionsClientOutputs:
  supported_extensions = {
      'appid', 'txAuthSimple', 'txAuthGeneric', 'authnSel', 'exts', 'uvi',
      'loc', 'uvm', 'biometricPerfBounds'
  }

  unsupported_extensions = set(
      extensions.keys()).difference(supported_extensions)

  if unsupported_extensions:
    raise ValidationError('Found unsupported extensions {}'.format(
        str(unsupported_extensions)))

  appid = extensions.get('appid')
  tx_auth_simple = extensions.get('txAuthSimple')
  tx_auth_generic = extensions.get('txAuthGeneric')
  authn_sel = extensions.get('authnSel')
  exts = extensions.get('exts')
  uvi = extensions.get('uvi')
  loc = extensions.get('loc')
  uvm = extensions.get('uvm')
  biometric_perf_bounds = extensions.get('biometricPerfBounds')

  if appid is not None:
    if type(appid) is not bool:
      raise ValidationError('appid extension client output should be a bool')

  if tx_auth_simple is not None:
    if type(tx_auth_simple) is not bool:
      raise ValidationError(
          'tx_auth_simple extension client output should be a str')

  if tx_auth_generic is not None:
    if type(tx_auth_generic) is not bytes:
      raise ValidationError(
          'tx_auth_generic extension client output should be bytes')

  if authn_sel is not None:
    if type(authn_sel) is not bool:
      raise ValidationError('authn_sel extension client output should be bool')

  if exts is not None:
    if type(exts) not in (list, tuple):
      raise ValidationError('exts extension client output should be list')

    for i, e in enumerate(exts):
      if type(e) is not str:
        raise ValidationError(
            'exts[{0}] extension client output should be str'.format(i))

  if uvi is not None:
    if type(uvi) is not bytes:
      raise ValidationError('uvi extension client output should be bytes')

  if loc is not None:
    if type(loc) is not dict:
      raise ValidationError('loc extension client output should be dict')

    if any(type(x) is not float for x in loc.values()):
      raise ValidationError('Coordinate value in loc extension must be float')

    supported_cvalues = {
        'latitude', 'longitude', 'altitude', 'accuracy', 'altitudeAccuracy',
        'heading', 'speed'
    }

    unsupported_cvalues = set(loc.keys()).difference(supported_cvalues)
    if unsupported_cvalues:
      raise ValidationError('Found unsupported loc key values {}'.format(
          str(unsupported_cvalues)))

    loc = Coordinates(latitude=loc.get('latitude'),
                      longitude=loc.get('longitude'),
                      altitude=loc.get('altitude'),
                      accuracy=loc.get('accuracy'),
                      altitude_accuracy=loc.get('altitudeAccuracy'),
                      heading=loc.get('heading'),
                      speed=loc.get('speed'))

  if uvm is not None:
    if type(uvm) is not list:
      raise ValidationError('uvm extension client output should be list')

    for i, uvm_entry in enumerate(uvm):
      if type(uvm_entry) is not list:
        raise ValidationError(
            'uvm[{0}] extension client output should be list'.format(i))

      for j, v in enumerate(uvm_entry):
        if type(v) is not str:
          raise ValidationError(
              'uvm[{0}][{1}] extension client output should be str'.format(
                  i, j))

  if biometric_perf_bounds is not None:
    if type(biometric_perf_bounds) is not bool:
      raise ValidationError(
          'biometric_perf_bounds extension client output should be bool')

  return AuthenticationExtensionsClientOutputs(
      appid=appid,
      tx_auth_simple=tx_auth_simple,
      tx_auth_generic=tx_auth_generic,
      authn_sel=authn_sel,
      exts=exts,
      uvi=uvi,
      loc=loc,
      uvm=uvm,
      biometric_perf_bounds=biometric_perf_bounds,
  )


def parse_attestation_statement_alg(
    att_stmt: dict
) -> Union[COSEAlgorithmIdentifier.Name, COSEAlgorithmIdentifier.Value]:
  alg = parse_dictionary_field('alg', (int, str), att_stmt)

  try:
    alg = COSEAlgorithmIdentifier(alg)  # type: ignore
  except KeyError:
    raise ValidationError('Invalid algorithm identifier {}'.format(alg))
  return alg


def parse_attestation_statement_x5c(att_stmt: dict) -> Sequence[bytes]:
  x5c = parse_dictionary_field('x5c', (list, tuple), att_stmt)

  for i, e in enumerate(x5c):
    if type(e) is not bytes:
      raise ValidationError('x5c[{}] must be a byte string'.format(i))
  return x5c


def parse_packed_attestation_statement(
    att_stmt: dict) -> PackedAttestationStatement:
  supported_keys = {'alg', 'sig'}
  alg = parse_attestation_statement_alg(att_stmt)
  sig = parse_dictionary_field('sig', bytes, att_stmt)

  if 'x5c' in att_stmt:
    supported_keys.add('x5c')
    check_unsupported_keys(supported_keys, att_stmt)
    x5c = parse_attestation_statement_x5c(att_stmt)
    return PackedX509AttestationStatement(alg=alg, sig=sig, x5c=x5c)

  if 'ecdaaKeyId' in att_stmt:
    supported_keys.add('ecdaaKeyId')
    check_unsupported_keys(supported_keys, att_stmt)
    ecdaa_key_id = parse_dictionary_field('ecdaaKeyId', bytes, att_stmt)
    return PackedECDAAAttestationStatement(alg=alg,
                                           sig=sig,
                                           ecdaa_key_id=ecdaa_key_id)

  check_unsupported_keys(supported_keys, att_stmt)
  return PackedAttestationStatement(alg=alg, sig=sig)


def parse_tpm_attestation_statement(att_stmt: dict) -> TPMAttestationStatement:
  supported_keys = {'alg', 'sig', 'ver', 'certInfo', 'pubArea'}
  alg = parse_attestation_statement_alg(att_stmt)
  sig = parse_dictionary_field('sig', bytes, att_stmt)
  ver = parse_dictionary_field('ver', str, att_stmt)
  cert_info = parse_dictionary_field('certInfo', bytes, att_stmt)
  pub_area = parse_dictionary_field('pubArea', bytes, att_stmt)

  if 'x5c' in att_stmt:
    supported_keys.add('x5c')
    check_unsupported_keys(supported_keys, att_stmt)
    x5c = parse_attestation_statement_x5c(att_stmt)
    return TPMX509AttestationStatement(alg=alg,
                                       sig=sig,
                                       ver=ver,
                                       cert_info=cert_info,
                                       pub_area=pub_area,
                                       x5c=x5c)

  if 'ecdaaKeyId' in att_stmt:
    supported_keys.add('ecdaaKeyId')
    check_unsupported_keys(supported_keys, att_stmt)
    ecdaa_key_id = parse_dictionary_field('ecdaaKeyId', bytes, att_stmt)
    return TPMECDAAAttestationStatement(alg=alg,
                                        sig=sig,
                                        ver=ver,
                                        cert_info=cert_info,
                                        pub_area=pub_area,
                                        ecdaa_key_id=ecdaa_key_id)

  check_unsupported_keys(supported_keys, att_stmt)
  return TPMAttestationStatement(alg=alg,
                                 sig=sig,
                                 ver=ver,
                                 cert_info=cert_info,
                                 pub_area=pub_area)


def parse_android_key_attestation_statement(
    att_stmt: dict) -> AndroidKeyAttestationStatement:
  supported_keys = {'alg', 'sig', 'x5c'}
  alg = parse_attestation_statement_alg(att_stmt)
  sig = parse_dictionary_field('sig', bytes, att_stmt)
  x5c = parse_attestation_statement_x5c(att_stmt)
  check_unsupported_keys(supported_keys, att_stmt)
  return AndroidKeyAttestationStatement(alg=alg, sig=sig, x5c=x5c)


def parse_android_safetynet_attestation_statement(
    att_stmt: dict) -> AndroidSafetyNetAttestationStatement:
  supported_keys = {'ver', 'response'}
  ver = parse_dictionary_field('ver', str, att_stmt)
  response = parse_dictionary_field('response', bytes, att_stmt)
  check_unsupported_keys(supported_keys, att_stmt)
  return AndroidSafetyNetAttestationStatement(ver=ver, response=response)


def parse_fido_u2f_attestation_statement(
    att_stmt: dict) -> FIDOU2FAttestationStatement:
  supported_keys = {'sig', 'x5c'}
  sig = parse_dictionary_field('sig', bytes, att_stmt)
  x5c = parse_attestation_statement_x5c(att_stmt)
  check_unsupported_keys(supported_keys, att_stmt)
  return FIDOU2FAttestationStatement(sig=sig, x5c=x5c)


def parse_none_attestation_statement(
    att_stmt: dict) -> NoneAttestationStatement:
  check_unsupported_keys(set(), att_stmt)
  return NoneAttestationStatement()


class AttestationStatementParser(Enum):
  PACKED = parse_packed_attestation_statement
  TPM = parse_tpm_attestation_statement
  ANDROID_KEY = parse_android_key_attestation_statement
  ANDROID_SAFETYNET = parse_android_safetynet_attestation_statement
  FIDO_U2F = parse_fido_u2f_attestation_statement
  NONE = parse_none_attestation_statement


def parse_client_data(
    client_data_JSON: bytes) -> Optional[CollectedClientData]:
  try:
    client_data_text = client_data_JSON.decode('utf-8')
    client_data = json.loads(client_data_text)
  except (UnicodeDecodeError, json.JSONDecodeError):
    raise DecodingError('Could not decode the client data JSON')

  type_ = client_data.get('type')
  challenge = client_data.get('challenge')
  origin = client_data.get('origin')

  if not all(isinstance(x, str) for x in (type_, challenge, origin)):
    raise ValidationError('Invalid client data parsed')

  token_binding_data = client_data.get('tokenBinding')
  token_binding = None
  if token_binding_data is not None:
    token_binding_status = token_binding_data.get('status')
    token_binding_id = token_binding_data.get('id')

    if token_binding_status == TokenBindingStatus.PRESENT and (token_binding_id
                                                               is None):
      raise TokenBindingError(
          'Token Binding must contain an id if status is {}'.format(
              TokenBindingStatus.PRESENT))

    token_binding = TokenBinding(status=token_binding_status,
                                 id=token_binding_id)

  return CollectedClientData(type=type_,
                             challenge=challenge,
                             origin=origin,
                             token_binding=token_binding)


def parse_cose_key(
    credential_public_key: Union[dict, bytes]) -> CredentialPublicKey:
  if type(credential_public_key) is bytes:
    credential_public_key = cbor.loads(credential_public_key)

  cose_key_type = cast(Union[COSEKeyType.Name, COSEKeyType.Value],
                       COSEKeyType(credential_public_key[1]))  # type: ignore

  try:
    cpk_parser = getattr(CredentialPublicKeyParser, cose_key_type.name)
  except AttributeError:
    raise ValidationError('Parser not supported for key type {}'.format(
        cose_key_type.name))

  return cpk_parser(credential_public_key)


def parse_authenticator_data(
    auth_data: bytes,
    fmt: Optional[AttestationStatementFormatIdentifier] = None
) -> AuthenticatorData:
  if len(auth_data) < 37:
    raise ValidationError('Attestation auth data must be at least 35 bytes')

  rp_id_hash = auth_data[:32]
  flags = auth_data[32]
  signature_counter_bytes = auth_data[33:37]
  signature_counter_uint32, = struct.unpack('>I', signature_counter_bytes)

  attested_credential_data = None
  aeci = None
  if len(auth_data) >= 53:
    aaguid = auth_data[37:53]
    credential_id_length_bytes = auth_data[53:55]
    credential_id_length_uint16, = struct.unpack('>H',
                                                 credential_id_length_bytes)
    credential_id = auth_data[55:(55 + credential_id_length_uint16)]
    credential_public_key_bytes = auth_data[(55 +
                                             credential_id_length_uint16):]

    try:
      credential_public_key, bytes_read = cbor.cbor._loads(
          io.BytesIO(credential_public_key_bytes))
    except ValueError:
      raise DecodingError('Could not decode the credential public key CBOR')

    if type(credential_public_key) is not dict:
      raise ValidationError('Credential public key must be a dictionary')

    cpk = parse_cose_key(credential_public_key)
    validate(cpk)

    extension_bytes = credential_public_key_bytes[bytes_read:]

    if extension_bytes:
      try:
        extensions = cbor.loads(extension_bytes)
      except ValueError:
        raise DecodingError('Could not decode the extensions CBOR')
      aeci = parse_extensions(extensions)

    attested_credential_data = AttestedCredentialData(
        aaguid=aaguid,
        credential_id_length=credential_id_length_uint16,
        credential_id=credential_id,
        credential_public_key=cpk,
    )

  return AuthenticatorData(
      rp_id_hash=rp_id_hash,
      flags=flags,
      sign_count=signature_counter_uint32,
      attested_credential_data=attested_credential_data,
      extensions=aeci,
  )


def parse_attestation(
    attestation_object: bytes) -> Tuple[AttestationObject, dict]:
  try:
    attestation_object_data = cbor.loads(attestation_object)
  except ValueError:
    raise DecodingError('Could not decode the attestation object CBOR')

  try:
    auth_data = attestation_object_data['authData']
    fmt = attestation_object_data['fmt']
    att_stmt = attestation_object_data['attStmt']

    if type(fmt) is not str:
      raise ValidationError('fmt must be a text string')

    try:
      asfi = AttestationStatementFormatIdentifier(fmt)
    except KeyError:
      raise ValidationError('Invalid attestation statement format identifier')

    if type(auth_data) is not bytes:
      raise ValidationError('Attestation auth data should be bytes')

    authenticator_data = parse_authenticator_data(auth_data, asfi)
  except KeyError as e:
    raise ValidationError('Missing {} in attestation'.format(str(e)))

  if type(att_stmt) is not dict:
    raise ValidationError('attStmt must be a dictionary')

  try:
    as_parser = getattr(AttestationStatementParser, asfi.name)
    attestation_statement = as_parser(att_stmt)
  except AttributeError:
    raise ValidationError('Unsupported attestation statement {}'.format(
        asfi.name))

  return AttestationObject(
      auth_data=authenticator_data,
      fmt=asfi,
      att_stmt=attestation_statement,
  ), attestation_object_data



import hashlib
import json
from typing import Any, Optional, Sequence, Set, cast

import cbor
import cryptography
from cryptography.hazmat.primitives.asymmetric.ec import ECDSA
from cryptography.hazmat.primitives.hashes import SHA256

from webauthn_rp.attesters import attest
from webauthn_rp.converters import cryptography_public_key, jsonify
from webauthn_rp.errors import (AuthenticationError, DecodingError,
                                IntegrityError, NotFoundError, ParseError,
                                RegistrationError, SignatureCountError,
                                ValidationError, VerificationError)
from webauthn_rp.parsers import (parse_attestation, parse_authenticator_data,
                                 parse_client_data)
from webauthn_rp.registrars import CredentialsRegistrar
from webauthn_rp.types import (
    AuthenticatorAssertionResponse, AuthenticatorAttestationResponse,
    AuthenticatorDataFlag, CredentialCreationOptions, CredentialRequestOptions,
    ExtensionIdentifier, PublicKeyCredential, PublicKeyCredentialDescriptor,
    PublicKeyCredentialRpEntity, PublicKeyCredentialUserEntity, TokenBinding)
from webauthn_rp.utils import extract_origin, url_base64_decode
from webauthn_rp.verifiers import verify


class CredentialsBackend:
  def __init__(self, registrar: CredentialsRegistrar):
    self.registrar = registrar

  def handle_creation_options(self, *, options: CredentialCreationOptions):
    if not self.registrar.register_creation_options(options):
      raise RegistrationError('Failed to register creation options')

  def handle_request_options(self, *, options: CredentialRequestOptions):
    if not self.registrar.register_request_options(options):
      raise RegistrationError('Failed to register request options')

  def handle_credential_creation(
      self,
      *,
      credential: PublicKeyCredential,
      user: PublicKeyCredentialUserEntity,
      rp: PublicKeyCredentialRpEntity,
      expected_challenge: bytes,
      token_binding: Optional[TokenBinding] = None,
      require_user_verification: bool = False,
      expected_extensions: Optional[Set[ExtensionIdentifier]] = None):
    response = cast(AuthenticatorAttestationResponse, credential.response)
    collected_client_data = parse_client_data(response.client_data_JSON)
    if collected_client_data is None:
      raise ParseError('Could not parse the client data dictionary')

    if collected_client_data.type != 'webauthn.create':
      raise ValidationError('Invalid client data type {}'.format(
          collected_client_data.type))

    try:
      challenge = url_base64_decode(collected_client_data.challenge)
      if challenge != expected_challenge:
        raise IntegrityError(
            'Given and expected challenge byte strings don\'t match')
    except ValueError:
      raise DecodingError('Failed to decode the base64 encoded challenge')

    rp_origin = extract_origin(collected_client_data.origin)
    if rp_origin != rp.id:
      raise IntegrityError(
          'Given ({0}) and expected ({1}) RP origin don\'t match'.format(
              rp_origin, rp.id))

    if token_binding is not None:
      if collected_client_data.token_binding is None:
        raise IntegrityError('Expecting Token Binding in client data')

      if token_binding.status != collected_client_data.token_binding.status:
        raise IntegrityError(('Given ({0}) and expected ({1}) Token Binding '
                              'statuses dont\'t match').format(
                                  collected_client_data.token_binding.status,
                                  token_binding.status))

      if token_binding.id != collected_client_data.token_binding.id:
        raise IntegrityError(('Given ({0}) and expected ({1}) Token Binding '
                              'ids dont\'t match').format(
                                  collected_client_data.token_binding.id,
                                  token_binding.id))
    elif collected_client_data.token_binding is not None:
      raise IntegrityError('Unexpected Token Binding in client data')

    client_data_JSON_hash = hashlib.sha256(response.client_data_JSON).digest()
    rp_id_hash = hashlib.sha256(rp.id.encode('utf-8')).digest()

    attestation, raw_att_obj = parse_attestation(response.attestation_object)
    if attestation.auth_data.rp_id_hash != rp_id_hash:
      raise IntegrityError('RP ID hash does not match')

    if not (AuthenticatorDataFlag.UP.value & attestation.auth_data.flags):
      raise ValidationError('User present flag must be set')

    if require_user_verification:
      if not (AuthenticatorDataFlag.UV.value & attestation.auth_data.flags):
        raise ValidationError('User verification flag must be set')

    if expected_extensions is not None:
      for e in expected_extensions:
        if not hasattr(attestation.auth_data.extensions, e.key) or getattr(
            attestation.auth_data.extensions, e.key) is None:
          raise ValidationError('Missing extension {}'.format(e.value))

    assert attestation.auth_data is not None
    assert attestation.auth_data.attested_credential_data is not None

    att_type, trusted_path = attest(
        attestation.att_stmt,
        attestation.auth_data.attested_credential_data.credential_public_key,
        raw_att_obj['authData'], client_data_JSON_hash)

    crypto_pk = cryptography_public_key(
        attestation.auth_data.attested_credential_data.credential_public_key)

    if not self.registrar.register_credential_creation(
        credential=credential,
        att=attestation,
        att_type=att_type,
        user=user,
        rp=rp,
        trusted_path=trusted_path,
    ):
      raise RegistrationError('Failed to register credential creation')

  def handle_credential_request(
      self,
      *,
      credential: PublicKeyCredential,
      expected_challenge: bytes,
      rp: Optional[PublicKeyCredentialRpEntity] = None,
      user: Optional[PublicKeyCredentialUserEntity] = None,
      allow_credentials: Optional[
          Sequence[PublicKeyCredentialDescriptor]] = None,
      token_binding: Optional[TokenBinding] = None,
      require_user_verification: bool = False,
      expected_extensions: Optional[Set[ExtensionIdentifier]] = None,
      ignore_clone_error: bool = False):
    response = cast(AuthenticatorAssertionResponse, credential.response)
    if allow_credentials is not None:
      allowed = False
      for cred in allow_credentials:
        if cred.id == credential.raw_id:
          allowed = True
          break

      if not allowed:
        raise ValidationError('Provided credential is not allowed')

    credential_data = self.registrar.get_credential_data(credential.raw_id)

    if credential_data is None:
      raise NotFoundError('Could not get credential data')

    registered_user = credential_data.user_entity
    if registered_user is None:
      if user is None:
        raise NotFoundError('Could not find a user with the credential')
    elif user is not None:
      if registered_user.id != user.id:
        raise ValidationError(
            'Registered user and provided user ids do not match')
    else:
      user = registered_user

    registered_rp = credential_data.rp_entity
    if registered_rp is None:
      if rp is None:
        raise NotFoundError(
            ('Could not find a registered rp with the credential '
             'and one was not provided for verification'))
    elif rp is not None:
      if registered_rp.id != rp.id:
        raise ValidationError('Registered rp and provided rp ids do not match')
    else:
      rp = registered_rp

    if user is not None:
      if response.user_handle is not None:
        if response.user_handle != user.id:
          raise ValidationError('User handle doesn\'t match user id')
    else:
      if response.user_handle is None:
        raise ValidationError(('User cannot be verified, must provide one or '
                               'there must be a user handle'))

      valid_credential = self.registrar.check_user_owns_credential(
          response.user_handle, credential.raw_id)

      if not valid_credential:
        raise ValidationError('User does not own credential')

    collected_client_data = parse_client_data(response.client_data_JSON)
    if collected_client_data is None:
      raise ParseError('Could not parse the client data dictionary')

    if collected_client_data.type != 'webauthn.get':
      raise ValidationError('Invalid client data type {}'.format(
          collected_client_data.type))

    try:
      challenge = url_base64_decode(collected_client_data.challenge)
      if challenge != expected_challenge:
        raise IntegrityError(
            'Given and expected challenge byte strings don\'t match')
    except ValueError:
      raise DecodingError('Failed to decode the base64 encoded challenge')

    rp_origin = extract_origin(collected_client_data.origin)
    if rp_origin != rp.id:
      raise IntegrityError(
          'Given ({0}) and expected ({1}) RP origin don\'t match'.format(
              rp_origin, rp.id))

    if token_binding is not None:
      if collected_client_data.token_binding is None:
        raise IntegrityError('Expecting Token Binding in client data')

      if token_binding.status != collected_client_data.token_binding.status:
        raise IntegrityError(('Given ({0}) and expected ({1}) Token Binding '
                              'statuses dont\'t match').format(
                                  collected_client_data.token_binding.status,
                                  token_binding.status))

      if token_binding.id != collected_client_data.token_binding.id:
        raise IntegrityError(('Given ({0}) and expected ({1}) Token Binding '
                              'ids dont\'t match').format(
                                  collected_client_data.token_binding.id,
                                  token_binding.id))
    elif collected_client_data.token_binding is not None:
      raise IntegrityError('Unexpected Token Binding in client data')

    client_data_JSON_hash = hashlib.sha256(response.client_data_JSON).digest()
    rp_id_hash = hashlib.sha256(rp.id.encode('utf-8')).digest()

    auth_data = parse_authenticator_data(response.authenticator_data)

    if auth_data.rp_id_hash != rp_id_hash:
      raise IntegrityError('RP ID hash does not match')

    if not (AuthenticatorDataFlag.UP.value & auth_data.flags):
      raise ValidationError('User present flag must be set')

    if require_user_verification:
      if not (AuthenticatorDataFlag.UV.value & auth_data.flags):
        raise ValidationError('User verification flag must be set')

    if expected_extensions is not None:
      for e in expected_extensions:
        if not hasattr(auth_data.extensions,
                       e.key) or getattr(auth_data.extensions, e.key) is None:
          raise ValidationError('Missing extension {}'.format(e.value))

    verification_data = response.authenticator_data + client_data_JSON_hash
    verify(credential_data.credential_public_key, response.signature,
           verification_data)

    registered_sign_count = credential_data.signature_count

    if auth_data.sign_count != 0 or registered_sign_count != 0:
      if auth_data.sign_count <= registered_sign_count:
        if not ignore_clone_error:
          raise SignatureCountError('Detected a possible authenticator clone')

    if not self.registrar.register_credential_request(
        credential=credential,
        authenticator_data=auth_data,
        user=user,
        rp=rp,
    ):
      raise RegistrationError('Failed to register credential request')



EC2_P_256_NUMBER_LENGTH = 32
EC2_P_384_NUMBER_LENGTH = 48
EC2_P_521_NUMBER_LENGTH = 66
OKP_ED25519_NUMBER_LENGTH = 32
OKP_ED448_NUMBER_LENGTH = 57

# https://android.googlesource.com/platform/hardware/libhardware/+/master/include/hardware/keymaster_defs.h
KM_ORIGIN_GENERATED = 0
KM_PURPOSE_SIGN = 2



from pyasn1.type.namedtype import NamedType, NamedTypes, OptionalNamedType
from pyasn1.type.namedval import NamedValues
from pyasn1.type.tag import Tag, tagClassContext, tagFormatSimple
from pyasn1.type.univ import (Boolean, Enumerated, Integer, Null, OctetString,
                              Sequence, SetOf)


class SecurityLevel(Enumerated):
  componentType = NamedValues(Software=0, TrustedEnvironment=1, StrongBox=2)


class VerifiedBootState(Enumerated):
  componentType = NamedValues(Verified=0, SelfSigned=1, Unverified=2, Failed=3)


class RootOfTrust(Sequence):
  componentType = NamedTypes(
      NamedType('verifiedBootKey', OctetString()),
      NamedType('deviceLocked', Boolean()),
      NamedType('verifiedBootState', VerifiedBootState()),
      NamedType('verifiedBootHash', OctetString()),
  )


class AuthorizationList(Sequence):
  componentType = NamedTypes(
      OptionalNamedType(
          'purpose',
          SetOf(Integer()).subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 1))),
      OptionalNamedType(
          'algorithm',
          Integer().subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 2))),
      OptionalNamedType(
          'keySize',
          Integer().subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 3))),
      OptionalNamedType(
          'digest',
          SetOf(Integer()).subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 5))),
      OptionalNamedType(
          'padding',
          SetOf(Integer()).subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 6))),
      OptionalNamedType(
          'ecCurve',
          Integer().subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 10))),
      OptionalNamedType(
          'rsaPublicExponent',
          Integer().subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 200))),
      OptionalNamedType(
          'rollbackResistance',
          Null().subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 303))),
      OptionalNamedType(
          'activeDateTime',
          Integer().subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 400))),
      OptionalNamedType(
          'originationExpireDateTime',
          Integer().subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 401))),
      OptionalNamedType(
          'usageExpireDateTime',
          Integer().subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 402))),
      OptionalNamedType(
          'noAuthRequired',
          Null().subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 503))),
      OptionalNamedType(
          'userAuthType',
          Integer().subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 504))),
      OptionalNamedType(
          'authTimeout',
          Integer().subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 505))),
      OptionalNamedType(
          'allowWhileOnBody',
          Null().subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 506))),
      OptionalNamedType(
          'trustedUserPresenceRequired',
          Null().subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 507))),
      OptionalNamedType(
          'trustedConfirmationRequired',
          Null().subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 508))),
      OptionalNamedType(
          'unlockedDeviceRequired',
          Null().subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 509))),
      OptionalNamedType(
          'allApplications',
          Null().subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 600))),
      OptionalNamedType(
          'applicationId',
          OctetString().subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 601))),
      OptionalNamedType(
          'creationDateTime',
          Integer().subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 701))),
      OptionalNamedType(
          'origin',
          Integer().subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 702))),
      OptionalNamedType(
          'rollbackResistant',
          Null().subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 703))),
      OptionalNamedType(
          'rootOfTrust',
          RootOfTrust().subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 704))),
      OptionalNamedType(
          'osVersion',
          Integer().subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 705))),
      OptionalNamedType(
          'osPatchLevel',
          Integer().subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 706))),
      OptionalNamedType(
          'attestationApplicationId',
          OctetString().subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 709))),
      OptionalNamedType(
          'attestationIdBrand',
          OctetString().subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 710))),
      OptionalNamedType(
          'attestationIdDevice',
          OctetString().subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 711))),
      OptionalNamedType(
          'attestationIdProduct',
          OctetString().subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 712))),
      OptionalNamedType(
          'attestationIdSerial',
          OctetString().subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 713))),
      OptionalNamedType(
          'attestationIdImei',
          OctetString().subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 714))),
      OptionalNamedType(
          'attestationIdMeid',
          OctetString().subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 715))),
      OptionalNamedType(
          'attestationIdManufacturer',
          OctetString().subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 716))),
      OptionalNamedType(
          'attestationIdModel',
          OctetString().subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 717))),
      OptionalNamedType(
          'vendorPatchLevel',
          Integer().subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 718))),
      OptionalNamedType(
          'bootPatchLevel',
          Integer().subtype(
              explicitTag=Tag(tagClassContext, tagFormatSimple, 719))),
  )


class KeyDescription(Sequence):
  componentType = NamedTypes(
      NamedType('attestationVersion', Integer()),
      NamedType('attestationSecurityLevel', SecurityLevel()),
      NamedType('keymasterVersion', Integer()),
      NamedType('keymasterSecurityLevel', SecurityLevel()),
      NamedType('attestationChallenge', OctetString()),
      NamedType('uniqueId', OctetString()),
      NamedType('softwareEnforced', AuthorizationList()),
      NamedType('teeEnforced', AuthorizationList()),
  )



from distutils.core import setup
from os.path import dirname, join

import setuptools


def get_version():
  with open(join(dirname(__file__), 'webauthn_rp/__init__.py'), 'r') as f:
    for line in f.read().splitlines():
      if line.startswith('__version__'):
        delim = '"' if '"' in line else "'"
        return line.split(delim)[1]
    else:
      raise RuntimeError("Unable to find version string.")


with open('requirements.txt') as rt:
  requirements_txt = rt.read()

with open('README.md') as rt:
  readme_txt = rt.read()

install_requires = [
    line.strip() for line in requirements_txt.split()
    if (not line.startswith('#'))
]

setup_requirements = [
    'setuptools>=38.6',
    'twine>=1.11',
    'wheel>=0.31',
]

test_requirements = [
    'pytest~=4.4',
    'pytest-cov~=2.7',
    'codecov~=2.0',
    'coverage~=4.5',
    'yapf~=0.27',
    'mypy~=0.701',
    'isort~=4.3',
]

setup(name='webauthn-rp',
      version=get_version(),
      description='Web Authentication Relying Party Library',
      author='enceladus-rex',
      packages=['webauthn_rp'],
      url='https://github.com/enceladus-rex/webauthn-rp',
      long_description=readme_txt,
      long_description_content_type='text/markdown',
      classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: MIT License",
          "Operating System :: OS Independent",
      ],
      test_suite='tests',
      install_requires=install_requires,
      setup_requirements=setup_requirements,
      test_requirements=test_requirements,
      extras_require={
          'test': test_requirements,
      })



from typing import Sequence

from webauthn_rp.types import (
    AttestationConveyancePreference, AuthenticationExtensionsClientInputs,
    AuthenticatorSelectionCriteria, CredentialCreationOptions,
    CredentialRequestOptions, PublicKeyCredentialCreationOptions,
    PublicKeyCredentialDescriptor, PublicKeyCredentialParameters,
    PublicKeyCredentialRpEntity, PublicKeyCredentialUserEntity)


class CredentialCreationOptionsBuilder:
  def __init__(
      self,
      rp: PublicKeyCredentialRpEntity,
      pub_key_cred_params: Sequence[PublicKeyCredentialParameters],
      timeout: int,
      authenticator_selection: AuthenticatorSelectionCriteria,
      extensions: AuthenticationExtensionsClientInputs,
      attestation:
      AttestationConveyancePreference = AttestationConveyancePreference.NONE,
      exclude_credentials: Sequence[PublicKeyCredentialDescriptor] = None):
    self.rp = rp
    self.pub_key_cred_params = pub_key_cred_params
    self.timeout = timeout
    self.authenticator_selection = authenticator_selection
    self.extensions = extensions
    self.attestation = attestation
    self.exclude_credentials = exclude_credentials

  def build(self, user: PublicKeyCredentialUserEntity,
            challenge: bytes) -> CredentialCreationOptions:
    return CredentialCreationOptions(
        public_key=PublicKeyCredentialCreationOptions(
            rp=self.rp,
            user=user,
            challenge=challenge,
            pub_key_cred_params=self.pub_key_cred_params,
            timeout=self.timeout,
            authenticator_selection=self.authenticator_selection,
            extensions=self.extensions,
            attestation=self.attestation,
            exclude_credentials=self.exclude_credentials))



from functools import singledispatch
from typing import cast

import cryptography
import cryptography.exceptions
from cryptography.hazmat.primitives.asymmetric.ec import ECDSA
from cryptography.hazmat.primitives.hashes import SHA256, SHA384, SHA512

from webauthn_rp.converters import cryptography_public_key
from webauthn_rp.errors import (UnimplementedError, ValidationError,
                                VerificationError)
from webauthn_rp.types import (COSEAlgorithmIdentifier, CredentialPublicKey,
                               EC2CredentialPublicKey, EC2KeyType,
                               EC2PublicKey, OKPCredentialPublicKey,
                               OKPKeyType, OKPPublicKey)


@singledispatch
def verify(credential_public_key: CredentialPublicKey, signature: bytes,
           data: bytes):
  raise UnimplementedError('Must implement verification for {}'.format(
      str(type(credential_public_key))))


@verify.register(EC2CredentialPublicKey)
def verify_ec2_credential_public_key(
    credential_public_key: EC2CredentialPublicKey, signature: bytes,
    data: bytes):
  public_key = cast(EC2PublicKey,
                    cryptography_public_key(credential_public_key))
  if credential_public_key.alg is None:
    raise ValidationError('alg must not be None')

  alg_name = credential_public_key.alg.name
  alg_to_hash = {
      'ES256': SHA256,
      'ES384': SHA384,
      'ES512': SHA512,
  }

  hash_algorithm = None
  if alg_name in alg_to_hash:
    hash_algorithm = ECDSA(alg_to_hash[alg_name]())
  else:
    raise ValidationError('Unsupported EC2 alg {}'.format(alg_name))

  try:
    public_key.verify(signature, data, hash_algorithm)
  except cryptography.exceptions.InvalidSignature:
    raise VerificationError('EC2 verification failure')


@verify.register(OKPCredentialPublicKey)
def verify_okp_credential_public_key(
    credential_public_key: OKPCredentialPublicKey, signature: bytes,
    data: bytes):
  public_key = cast(OKPPublicKey,
                    cryptography_public_key(credential_public_key))

  try:
    public_key.verify(signature, data)
  except cryptography.exceptions.InvalidSignature:
    raise VerificationError('OKP verification failure')
